# This file is part of libkd.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
from astrometry.libkd import spherematch
import numpy as np
from time import time

N1 = 1000
N2 = 1000
D = 2
r = 0.02

x1 = np.random.rand(N1, D)
x2 = np.random.rand(N2, D)

t0 = time()
(inds,dists) = spherematch.match(x1, x2, r)
dt = time() - t0

print('spherematch.match: found', len(inds), 'pairs in', int(dt*1000.), 'ms')

order = np.argsort(inds[:,0]*N2 + inds[:,1])
inds = inds[order]
dists = dists[order]

t0 = time()
pairs = []
truedists = []
for i in range(N1):
    pt1 = x1[i,:]
    d2s = np.sum((x2 - pt1)**2, axis=1)
    good = np.where(d2s <= r**2)[0]
    for j in good:
        pairs.append((i, j))
        truedists.append(d2s[j])
dt = time() - t0
pairs = np.array(pairs)
truedists = np.sqrt(np.array(truedists))

print('naive            : found', len(pairs), 'pairs in', int(dt*1000.), 'ms')

order = np.argsort(pairs[:,0]*N2 + pairs[:,1])
pairs = pairs[order]

ok = np.array_equal(pairs, inds)
print('Indices equal:', ok)

ok = np.array_equal(truedists[order], dists.ravel())
print('Dists equal:', ok)


### Repeat for u64 trees

S = float(1<<63)
ux1 = (x1 * S).astype(np.uint64)
ux2 = (x2 * S).astype(np.uint64)

t0 = time()
(inds,dists) = spherematch.match(ux1, ux2, r*S)
dt = time() - t0

print('spherematch.match: found', len(inds), 'pairs in', int(dt*1000.), 'ms with uint64 trees')
order = np.argsort(inds[:,0]*N2 + inds[:,1])
inds = inds[order]
dists = dists[order]
ok = np.array_equal(pairs, inds)
print('Indices equal:', ok)

print('Build uint64 tree...')
kd = spherematch.tree_build(ux1)
kd.print()

data = kd.get_data(np.array([0,3,5]).astype(np.uint32))
assert(data.dtype == np.uint64)
print('Kd data:', data.dtype, data)

kd.write('kd-u64.fits')

kd2 = spherematch.tree_open('kd-u64.fits')
data2 = kd2.get_data(np.array([0,3,5]).astype(np.uint32))
assert(data2.dtype == np.uint64)
print('Kd data2:', data2.dtype, data2)
assert(np.all(data == data2))

del kd
del kd2

###



t0 = time()
(inds,dists) = spherematch.nearest(x1, x2, r)
dt = time() - t0

t0 = time()
inds = spherematch.match(x1, x2, r, indexlist=True)
dt = time() - t0

kd = spherematch.tree_build(x1)
kd.print()
R = spherematch.tree_search(kd, x2[0,:], 1.)
print('tree_search:', len(R), 'results')
spherematch.tree_close(kd)

kd2 = spherematch.tree_build(x2)
I,J,d = spherematch.trees_match(kd, kd2, 1.)
print('trees_match:', len(I), 'matches')

I,J,d = spherematch.trees_match(kd, kd2, 1., nearest=True)
print('trees_match:', len(I), 'matches (nearest)')

print('Kd bounding-box:', spherematch.tree_bbox(kd))

print('Kd bounding-box:', kd.bbox)

print('Kd data:', spherematch.tree_data(kd, np.array([0,3,5]).astype(np.uint32)))

print('Kd data:', kd.get_data(np.array([0,3,5]).astype(np.uint32)))

print('Permute:', spherematch.tree_permute(kd, np.array([3,5,7]).astype(np.int32)))

print('Permute:', kd.permute(np.array([0,99,199]).astype(np.int32)))

ra,dec = np.meshgrid(np.arange(0, 360), np.arange(-90, 91, 1))
ra1 = ra.ravel()
dec1 = dec.ravel()
rdkd1 = spherematch.tree_build_radec(ra1, dec1)
print('RdKd:', rdkd1.n, rdkd1.bbox)

ra2  = np.random.uniform(-10, 10, size=1000)
dec2 = np.random.uniform(-10, 10, size=1000)
rdkd2 = spherematch.tree_build_radec(ra2, dec2)

I = spherematch.tree_search_radec(rdkd1, ra2[0], dec2[0], 2.)
print('search_radec:', I)

I,J,d = spherematch.match_radec(ra1, dec1, ra2, dec2, 1.)
print('Matches:', len(I))

I,J,d = spherematch.match_radec(ra1, dec1, ra2, dec2, 1., nearest=True)
print('Nearest matches:', len(I))

I = spherematch.match_radec(ra1, dec1, ra2, dec2, 1.,
                            indexlist=True)
print('Index lists matches:', len(I))



# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
from __future__ import absolute_import
import numpy as np
from scipy.ndimage.filters import gaussian_filter

from astrometry.util.resample import resample_with_wcs, ResampleError

from .fields import radec_to_sdss_rcf
from .common import band_name, band_index, AsTransWrapper
from functools import reduce

def get_sdss_cutout(targetwcs, sdss, get_rawvals=False, bands='irg',
                    get_rawvals_only=False,
                    bandscales=dict(z=1.0, i=1.0, r=1.3, g=2.5)):

    rgbims = []

    ra,dec = targetwcs.radec_center()
    # in deg
    radius = targetwcs.radius()
    #print 'Target WCS radius is', radius, 'deg'
    H,W = targetwcs.get_height(), targetwcs.get_width()
    targetpixscale = targetwcs.pixel_scale()
    
    wlistfn = sdss.filenames.get('window_flist', 'window_flist.fits')
    rad2 = radius*60. + np.hypot(14., 10.)/2.
    #print 'Rad2 radius', rad2, 'arcmin'
    RCF = radec_to_sdss_rcf(ra, dec, tablefn=wlistfn, radius=rad2)

    # Drop rerun 157
    keepRCF = []
    for run,camcol,field,r,d in RCF:
        rr = sdss.get_rerun(run, field)
        #print 'Rerun:', rr
        if rr == '157':
            continue
        keepRCF.append((run,camcol,field))
    RCF = keepRCF
    print(len(RCF), 'run/camcol/fields in range')

    # size in SDSS pixels of the target image.
    sz = np.hypot(H, W)/2. * targetpixscale / 0.396
    print('SDSS sz:', sz)

    bandnums = [band_index(b) for b in bands]
    
    for bandnum,band in zip(bandnums, bands):
        targetim = np.zeros((H, W), np.float32)
        targetn  = np.zeros((H, W), np.uint8)

        for ifield,(run,camcol,field) in enumerate(RCF):
            
            fn = sdss.retrieve('frame', run, camcol, field, band)
            frame = sdss.readFrame(run, camcol, field, bandnum)

            h,w = frame.getImageShape()
            x,y = frame.astrans.radec_to_pixel(ra, dec)
            x,y = int(x), int(y)
            # add some margin for resampling
            sz2 = int(sz) + 5
            xlo = np.clip(x - sz2, 0, w)
            xhi = np.clip(x + sz2 + 1, 0, w)
            ylo = np.clip(y - sz2, 0, h)
            yhi = np.clip(y + sz2 + 1, 0, h)
            if xlo == xhi or ylo == yhi:
                continue
            stamp = frame.getImageSlice((slice(ylo, yhi), slice(xlo, xhi)))
            sh,sw = stamp.shape
            wcs = AsTransWrapper(frame.astrans, sw, sh, x0=xlo, y0=ylo)
            # FIXME -- allow nn resampling too
            try:
                Yo,Xo,Yi,Xi,[rim] = resample_with_wcs(targetwcs, wcs, [stamp], 3)
            except ResampleError:
                continue
            targetim[Yo,Xo] += rim
            targetn [Yo,Xo] += 1

        rgbims.append(targetim / targetn)

    if get_rawvals_only:
        return rgbims

    if get_rawvals:
        rawvals = [x.copy() for x in rgbims]

    r,g,b = rgbims

    r *= bandscales[bands[0]]
    g *= bandscales[bands[1]]
    b *= bandscales[bands[2]]
    
    # i
    #r *= 1.0
    # r
    #g *= 1.5
    #g *= 1.3
    # g
    #b *= 2.5
    m = -0.02
    r = np.maximum(0, r - m)
    g = np.maximum(0, g - m)
    b = np.maximum(0, b - m)
    I = (r+g+b)/3.
    alpha = 1.5
    Q = 20
    m2 = 0.
    fI = np.arcsinh(alpha * Q * (I - m2)) / np.sqrt(Q)
    I += (I == 0.) * 1e-6
    R = fI * r / I
    G = fI * g / I
    B = fI * b / I
    maxrgb = reduce(np.maximum, [R,G,B])
    J = (maxrgb > 1.)
    R[J] = R[J]/maxrgb[J]
    G[J] = G[J]/maxrgb[J]
    B[J] = B[J]/maxrgb[J]
    ss = 0.5
    RGBblur = np.clip(np.dstack([
        gaussian_filter(R, ss),
        gaussian_filter(G, ss),
        gaussian_filter(B, ss)]), 0., 1.)

    if get_rawvals:
        return RGBblur, rawvals
    return RGBblur
        


if __name__ == '__main__':
    import tempfile
    import matplotlib
    matplotlib.use('Agg')
    import pylab as plt

    from .dr10 import DR10
    from astrometry.util.util import Tan
    
    tempdir = tempfile.gettempdir()

    sdss = DR10(basedir=tempdir)
    sdss.saveUnzippedFiles(tempdir)

    W,H = 100, 100
    pixscale = 1.
    cd = pixscale / 3600.
    targetwcs = Tan(120., 10., W/2., H/2., -cd, 0., 0., cd, float(W), float(H))
    rgb = get_sdss_cutout(targetwcs, sdss)

    plt.clf()
    plt.imshow(rgb, interpolation='nearest', origin='lower')
    plt.savefig('cutout1.png')
    

    W,H = 3000, 3000
    pixscale = 0.5
    cd = pixscale / 3600.
    targetwcs = Tan(120., 10., W/2., H/2., -cd, 0., 0., cd, float(W), float(H))
    rgb = get_sdss_cutout(targetwcs, sdss)

    plt.clf()
    plt.imshow(rgb, interpolation='nearest', origin='lower')
    plt.savefig('cutout2.png')
    



# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
try:
    import pyfits
except ImportError:
    try:
        from astropy.io import fits as pyfits
    except ImportError:
        raise ImportError("Cannot import either pyfits or astropy.io.fits")
from optparse import OptionParser
from astrometry.util.fits import *

def match_to_wcs(matchfn, row, wcsfn, imgw=0, imgh=0):
    m = table_fields(matchfn)
    
    wcs = pyfits.PrimaryHDU()
    h = wcs.header

    hdrs = [
        ('CTYPE1', 'RA---TAN', None),
        ('CTYPE2', 'DEC--TAN', None),
        ('WCSAXES', 2, None),
        ('EQUINOX', 2000.0, 'Equatorial coordinates definition (yr)'),
        ('LONPOLE', 180.0, None),
        ('LATPOLE', 0.0, None),
        ('CUNIT1', 'deg', 'X pixel scale units'),
        ('CUNIT2', 'deg', 'Y pixel scale units'),
        ('CRVAL1', m.crval[row][0], 'RA  of reference point'),
        ('CRVAL2', m.crval[row][1], 'DEC of reference point'),
        ('CRPIX1', m.crpix[row][0], 'X reference pixel'),
        ('CRPIX2', m.crpix[row][1], 'Y reference pixel'),
        ('CD1_1', m.cd[row][0], 'Transformation matrix'),
        ('CD1_2', m.cd[row][1], None),
        ('CD2_1', m.cd[row][2], None),
        ('CD2_2', m.cd[row][3], None),
        ]
    if imgw:
        hdrs.append(('IMAGEW', imgw, 'Image width,  in pixels.'))
    if imgh:
        hdrs.append(('IMAGEH', imgh, 'Image height, in pixels.'))

    for (k,v,c) in hdrs:
        h.update(k, v, c)
    pyfits_writeto(wcs, wcsfn)


if __name__ == '__main__':
    parser = OptionParser(usage='%prog [options] <match-input-file> <wcs-output-file>')

    parser.add_option('-r', dest='row', help='Row of the match file to take (default 0)', type='int')
    parser.add_option('-W', dest='imgw', help='Image width', type='float')
    parser.add_option('-H', dest='imgh', help='Image height', type='float')

    parser.set_defaults(row=0, imgw=0, imgh=0)

    (options, args) = parser.parse_args()
    if len(args) != 2:
        parser.print_help()
        print()
        print('Need args <match-input-file> and <wcs-output-file>')
        sys.exit(-1)

    matchfn = args[0]
    wcsfn = args[1]

    match_to_wcs(matchfn, options.row, wcsfn, options.imgw, options.imgh)




# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
import sys
import os

def addpath():
    try:
        import astrometry
        from astrometry.util.shell import shell_escape
        from astrometry.util.filetype import filetype_short
    except ImportError:
        me = __file__
        path = os.path.realpath(me)
        utildir = os.path.dirname(path)
        assert(os.path.basename(utildir) == 'util')
        andir = os.path.dirname(utildir)
        if os.path.basename(andir) == 'astrometry':
            rootdir = os.path.dirname(andir)
            sys.path.insert(1, andir)
        else:
            # assume there's a symlink astrometry -> .
            rootdir = andir
        #sys.path += [rootdir]
        sys.path.insert(1, rootdir)



#! /usr/bin/env python3
# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function

import sys
try:
    import pyfits
except ImportError:
    try:
        from astropy.io import fits as pyfits
    except ImportError:
        raise ImportError("Cannot import either pyfits or astropy.io.fits")

from math import *
from numpy import *
from pylab import *
from scipy.ndimage.filters import *
from astrometry.util.fits import pyfits_writeto

def normalized_hough(x, y, imgw, imgh, rlo, rhi, tlo, thi, nr, nt):
    houghimg = zeros((nr, nt)).astype(int)
    tstep = (thi - tlo) / float(nt)
    rstep = (rhi - rlo) / float(nr)
    # For each point, accumulate into the Hough transform image...
    tt = tlo + (arange(nt) + 0.5) * tstep
    cost = cos(tt)
    sint = sin(tt)
    for (xi,yi) in zip(x, y):
        rr = xi * cost + yi * sint
        ri = floor((rr - rlo) / rstep).astype(int)
        I = (ri >= 0) * (ri < nr)
        houghimg[ri[I], I] += 1

    # Compute the approximate Hough normalization image
    houghnorm = zeros((nr, nt)).astype(float)
    rr = rlo + (arange(nr) + 0.5) * rstep
    for ti,t in enumerate(tt):
        (x0,x1,y0,y1) = clip_to_image(rr, t, imgw, imgh)
        dist = sqrt((x0 - x1)**2 + (y0 - y1)**2)
        houghnorm[:, ti] = dist
    # expected number of points: dist is the length of the slice,
    #   rstep is the width of the slice; len(x)/A is the source density.
    houghnorm *= (rstep * len(x) / (imgw*imgh))
    return (houghimg, houghnorm, rr, tt, rstep, tstep)

def clip_to_image(r, t, imgw, imgh):
    eps = 1e-9
    if abs(t) < eps or abs(t-pi) < eps:
        # near-vertical.
        s = (abs(t) < eps) and 1 or -1
        y0 = 0
        y1 = ((r*s >= 0) * (r*s < imgw)) * imgh
        x0 = x1 = clip(r, 0, imgw)
        return (x0, x1, y0, y1)
    m = -cos(t)/sin(t)
    b = r/sin(t)
    x0 = 0
    x1 = imgw
    y0 = clip(b + m*x0, 0, imgh)
    y1 = clip(b + m*x1, 0, imgh)
    x0 = clip((y0 - b) / m, 0, imgw)
    x1 = clip((y1 - b) / m, 0, imgw)
    y0 = clip(b + m*x0, 0, imgh)
    y1 = clip(b + m*x1, 0, imgh)
    return (x0, x1, y0, y1)

def removelines_general(infile, outfile, nt=180, nr=180, thresh1=2.,
                        thresh2=5., plots=False):
    p = pyfits.open(infile)
    xy = p[1].data
    hdr = p[1].header
    x = xy.field('X').copy()
    y = xy.field('Y').copy()

    imshowargs = { 'interpolation':'nearest', 'origin':'lower' }

    imgw = int(ceil(max(x) - min(x)))
    imgh = int(ceil(max(y) - min(y)))

    x -= min(x)
    y -= min(y)

    Rmax = sqrt(imgw**2 + imgh**2)
    Rmin = -Rmax

    (houghimg, houghnorm, rr, tt, rstep, tstep
     ) = normalized_hough(x, y, imgw, imgh, Rmin, Rmax, 0, pi, nr, nt)

    hnorm = houghimg / maximum(houghnorm, 1)

    if plots:
        clf()
        plot(x,y,'r.')
        savefig('xy.png')

        clf()
        imshow(houghimg, **imshowargs)
        xlabel('Theta')
        ylabel('Radius')
        colorbar()
        savefig('hough.png')

        clf()
        imshow(houghnorm, **imshowargs)
        xlabel('Theta')
        ylabel('Radius')
        colorbar()
        savefig('norm.png')

        clf()
        imshow(hnorm, **imshowargs)
        xlabel('Theta')
        ylabel('Radius')
        colorbar()
        savefig('hnorm.png')


    I = find(hnorm.ravel() >= thresh1)
    print('%i peaks are above the coarse threshold' % len(I))
    bestri = I / nt
    bestti = I % nt

    if plots:
        a=axis()
        for (ri,ti) in zip(bestri,bestti):
            plot([ti-2, ti-2, ti+2, ti+2, ti-2], [ri-2, ri+2, ri+2, ri-2, ri-2], 'r-')
            axis(a)
            savefig('zooms.png')

        clf()
        plot(x,y,'r.')
        for (ri,ti) in zip(bestri,bestti):
            (x0,x1,y0,y1) = clip_to_image(rr[ri], tt[ti], imgw, imgh)
            plot([x0,x1],[y0,y1], 'b-')
        savefig('xy2.png')

    # how big a search area around each peak?
    boxsize = 1
    # how much more finely to grid.
    finer = 3
    nr2 = (boxsize * 2)*finer + 1
    nt2 = nr2

    bestrt = []
    keep = array([True] * len(x))
    for (ri,ti) in zip(bestri,bestti):
        (subh, subnorm, subrr, subtt, subrstep, subtstep
         ) = normalized_hough(x, y, imgw, imgh,
                              rr[max(ri-boxsize, 0)], rr[min(ri+boxsize, nr-1)],
                              tt[max(ti-boxsize, 0)], tt[min(ti+boxsize, nt-1)],
                              nr2, nt2)
        #print '  median normalization:', median(subnorm)
        subhnorm = subh / maximum(subnorm,1)
        I = find((subhnorm).ravel() >= thresh2)
        for i in I:
            bestsubri = i / nt2
            bestsubti = i % nt2
            r = subrr[bestsubri]
            t = subtt[bestsubti]
            bestrt.append((r,t))
            #print '  (r=%.1f, t=%.1f): factor %.1f above expected' % (r, t*180/pi, subhnorm.ravel()[i])
            thisr = x * cos(t) + y * sin(t)
            keep *= (abs(thisr - r) > subrstep/2.)

    print('In finer grid: found %i peaks' % len(bestrt))

    if plots:
        clf()
        subplot(1,1,1)
        plot(x,y,'r.')
        for (r,t) in bestrt:
            (x0,x1,y0,y1) =  clip_to_image(r, t, imgw, imgh)
            plot([x0,x1],[y0,y1],'b-')
        savefig('xy3.png')

        clf()
        plot(x,y,'r.')
        plot(x[keep == False], y[keep == False], 'b.')
        savefig('xy4.png')

    p[1].data = p[1].data[keep]
    pyfits_writeto(p, outfile)
    return 0

def exact_hough_normalization():
    houghnorm = zeros((nr, nt)).astype(float)
    [xx,yy] = meshgrid(arange(imgw), arange(imgh))
    yyflat = yy.ravel()
    xxflat = xx.ravel()
    for ti in range(nt):
        print(ti)
        t = (ti+0.5) * tstep
        rr = xxflat * cos(t) + yyflat * sin(t)
        ri = floor((rr - Rmin) / rstep).astype(int)
        (counts, nil) = histogram(ri, range(0, nr+1))
        houghnorm[:, ti] += counts
    clf()
    imshow(houghnorm, **imshowargs)
    colorbar()
    savefig('houghnorm.png')
    


if __name__ == '__main__':
    args = sys.argv[1:]
    plots = False
    if '-p' in args:
        plots = True
        args.remove('-p')

    if len(args) != 2:
        print('Usage: %s [options] <input-file> <output-file>' % sys.argv[0])
        print('   [-p]: create plots')
        exit(-1)

    infile = args[0]
    outfile = args[1]
    rtncode = removelines_general(infile, outfile, plots=plots)
    sys.exit(rtncode)




# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
import matplotlib
matplotlib.use('Agg')

from math import floor, ceil
from matplotlib.ticker import FixedFormatter

from astrometry.util.util import *

import numpy as np
import pylab as plt

if __name__ == '__main__':
    #indexfn = '/Users/dstn/INDEXES/tycho-2/tycho-2811.index'
    #magname = 'MAG'
    
    # scp sesame3:/scratch/lustre/hsc/astrometry_net_data/sdss-dr8/index-110311000.fits /tmp
    #indexfn = '/tmp/index-110311000.fits'
    indexfn = '/tmp/index-110311003.fits'
    magname = 'r'

    index = index_load(indexfn, 0, None)
    print(index)

    print('Index', indexfn, 'has', index_nstars(index), 'stars and', end=' ')
    print(index_nquads(index), 'quads')

    quads = index.quads
    print('quads:', quads)

    starkd = index.starkd;
    print('star kdtree:', starkd)

    tagalong = startree_get_tagalong(starkd)
    print('tag-along:', tagalong)

    mags = startree_get_data_column(starkd, magname, None, startree_N(starkd))
    print('mags:', mags[:10], '...')

    print('mag range:', min(mags), max(mags))

    mn,mx = [],[]
    for i in range(index_nquads(index)):
        if i % 100000 == 0:
            print('quad', i, 'of', index_nquads(index))
        stars = quadfile_get_stars(quads, i)
        #print 'stars', stars
        qmags = [mags[j] for j in stars]
        #print 'mags', qmags
        mn.append(min(qmags))
        mx.append(max(qmags))

        #if i == 1000000:
        #    break
        
    mn = np.array(mn)
    mx = np.array(mx)
        
    H,xe,ye = np.histogram2d(mn, mx, bins=50)
    plt.clf()
    plt.imshow(np.log10(1+H.T), extent=[min(xe),max(xe),min(ye),max(ye)],
               origin='lower', interpolation='nearest')
    plt.xlabel('Min mag')
    plt.ylabel('Max mag')
    a = plt.axis()
    lo = int(floor(min(mn)))
    hi = int(ceil(max(mn)))
    for delta in range(5):
        plt.plot([lo, hi], [lo+delta, hi+delta], 'k-')
    plt.axis(a)
    plt.title('Range of mags in index %s' % index.indexname)

    mx = np.log10(1+H.max())
    tens = np.arange(mx)
    plt.colorbar(ticks=tens, format=FixedFormatter(['%i'%10**i for i in tens]))
    plt.savefig('magrange.png')


    H,xe,ye = np.histogram2d(mn, mx-mn, bins=50)
    plt.clf()
    plt.imshow(np.log10(1+H.T), extent=[min(xe),max(xe),min(ye),max(ye)],
               origin='lower', interpolation='nearest')
    plt.xlabel('Min mag')
    plt.ylabel('Mag range')
    plt.savefig('magrange2.png')



#! /usr/bin/env python3
# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
from __future__ import absolute_import
import matplotlib
matplotlib.use('Agg')
import os
import sys
import logging

from pylab import *
from optparse import OptionParser

from astrometry.util.removelines import hist_remove_lines
from astrometry.util.fits import pyfits_writeto

if __name__ == '__main__':
    try:
        try:
            import pyfits
        except ImportError:
            try:
                from astropy.io import fits as pyfits
            except ImportError:
                raise ImportError(
                    "Cannot import either pyfits or astropy.io.fits"
                )
        import astrometry
        from astrometry.util.shell import shell_escape
        from astrometry.util.filetype import filetype_short
    except ImportError:
        me = sys.argv[0]
        #print 'i am', me
        path = os.path.realpath(me)
        #print 'my real path is', path
        utildir = os.path.dirname(path)
        assert(os.path.basename(utildir) == 'util')
        andir = os.path.dirname(utildir)
        #assert(os.path.basename(andir) == 'astrometry')
        rootdir = os.path.dirname(andir)
        #print 'adding path', rootdir
        #sys.path += [rootdir]
        sys.path.insert(1, andir)
        sys.path.insert(2, rootdir)

import numpy
try:
    import pyfits
except ImportError:
    try:
        from astropy.io import fits as pyfits
    except ImportError:
            raise ImportError("Cannot import either pyfits or astropy.io.fits")
from numpy import *
from numpy.random import rand

'''
# Returns a numpy array of booleans: True for points that should be kept (are not part of lines)
if False:
    # def hist_remove_lines(x, binwidth, binoffset, nsig, plots=False):
    bins = -binoffset + arange(0, max(x)+binwidth, binwidth)
    (counts, thebins) = histogram(x, bins)

    # We're ignoring empty bins.
    occupied = nonzero(counts)[0]
    noccupied = len(occupied)
    #k = (counts[occupied] - 1)
    #mean = sum(k) / float(noccupied)
    k = counts[occupied]
    mean = sum(k) / ((max(x) - min(x)) / binwidth) * sqrt(2.)
    thresh = mean + nsig * sqrt(mean)

    if plots:
        hist(x, bins)
        axhline(mean)
        axhline(thresh, color='r')

    print 'mean', mean, 'thresh:', thresh, 'max:', max(k)

    #logpoisson = k*log(mean) - mean - array([sum(log(1 + arange(kk))) for kk in k])
    #uk = unique(k)
    #ulogpoisson = uk*log(mean) - mean - array([sum(1+arange(kk)) for kk in uk])
    #print
    #for (uuk,ull) in zip(uk,ulogpoisson):
    #    print uuk,ull

    #badbins = occupied[logpoisson < logcut]
    badbins = occupied[k > thresh]
    if len(badbins) == 0:
        return array([True] * len(x))

    badleft = bins[badbins]
    badright = badleft + binwidth

    badpoints = sum(array([(x >= L)*(x < R) for (L,R) in zip(badleft, badright)]), 0)
    return (badpoints == 0)
'''
#'''

def removelines(infile, outfile, xcol='X', ycol='Y', plots=False, cut=None, **kwargs):
    p = pyfits.open(infile)
    xy = p[1].data
    hdr = p[1].header
    x = xy.field(xcol)
    y = xy.field(ycol)

    NX = max(x) - min(x)
    NY = max(y) - min(y)
    nangle = int(ceil(sqrt(NX*NY)/4.))

    if cut is None:
        cut = 20

    if plots:
        clf()
        plot(x, y, 'r.')

    I = array([True]*len(x))
    for i,angle in enumerate(0.75 + linspace(0, pi/2., nangle, endpoint=False)):
        cost = cos(angle)
        sint = sin(angle)
        xx = x*cost     + y*sint
        yy = x*-sint + y*cost
        xx -= min(xx)
        yy -= min(yy)

        if plots:
            print()
            clf()
            subplot(2,2,1)
            plot(xx, yy, 'r.')
            subplot(2,2,3)


        #ix = hist_remove_lines(xx, 0.5, 0.5, 5, plots=plots)
        ix = hist_remove_lines(xx, 1, 0.5, logcut=-cut)

        if plots:
            subplot(2,2,4)

        #iy = hist_remove_lines(yy, 0.5, 0.5, 5, plots=plots)
        iy = hist_remove_lines(yy, 1, 0.5, logcut=-cut)

        I *= ix * iy

        removed = (ix * iy == False)
        if plots:
            if sum(removed):
                plot([min(x[removed]), max(x[removed])],
                     [min(y[removed]), max(y[removed])], 'k-', alpha=0.5)
            subplot(2,2,1)
            plot(xx[removed], yy[removed], 'b-', alpha=0.5)
            plot(xx[removed], yy[removed], 'b.')
            savefig('rot-%04i.png' % i)

        print('angle', angle, 'removed', (len(x) - sum(ix*iy)))

    xc = x[I]
    yc = y[I]
    print('removelines.py: Removed %i sources' % (len(x) - len(xc)))

    if plots:
        plot(xc, yc, 'o', mec='r', mfc='none')
        # axes('equal')
        savefig('after.png')

    p[1].header.add_history('This xylist was filtered by the "removelines_rotate.py" program')
    p[1].header.add_history('to remove lines of sources')
    p[1].header.update('REMLINEN', len(x) - len(xc), 'Number of sources removed by "removelines"')

    p[1].data = p[1].data[I]
    pyfits_writeto(p, outfile)

    return 0

if __name__ == '__main__':
    parser = OptionParser(usage='%prog <input-file> <output-file>')
    parser.add_option('-X', '--x-column', dest='xcol', help='X column name')
    parser.add_option('-Y', '--y-column', dest='ycol', help='Y column name')
    parser.set_defaults(xcol='X', ycol='Y')
    opt,args = parser.parse_args()
    if len(args) != 2:
        parser.print_help()
        sys.exit(0)

    kwargs = {}
    for k in 'xcol','ycol':
        kwargs[k] = getattr(opt, k)

    infile = args[0]
    outfile = args[1]
    rtncode = removelines(infile, outfile, **kwargs)
    sys.exit(rtncode)




# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
import matplotlib
matplotlib.use('Agg')
import pylab as plt
import numpy as np

from astrometry.util.util import *
from astrometry.util.starutil_numpy import *

def plot_healpix_boundaries(nside):
    for hp in range(12*nside**2):
        # walk healpix boundary in ra,dec
        rd = []
        dd = np.linspace(0, 1, 100)
        x = 0
        rd.extend([healpix_to_radecdeg(hp, nside, x, y)
                   for y in dd])
        y = 1
        rd.extend([healpix_to_radecdeg(hp, nside, x, y)
                   for x in dd])
        x = 1
        rd.extend([healpix_to_radecdeg(hp, nside, x, y)
                   for y in reversed(dd)])
        y = 0
        rd.extend([healpix_to_radecdeg(hp, nside, x, y)
                   for x in reversed(dd)])

        rd = np.array(rd)
        ra,dec = rd[:,0], rd[:,1]

        # Put label in center of healpix
        xyz = radectoxyz(ra, dec)
        xyz = np.mean(xyz, axis=0)
        rc,dc = xyztoradec(xyz)
        plt.text(rc, dc, '%i' % hp, color='b')

        # handle RA wrap-around
        if rc > 180:
            ra += (ra < 90)*360
        else:
            ra -= (ra > 270)*360
        plt.plot(ra, dec, 'b-')

    plt.axis([360,0,-90,90])


if __name__ == '__main__':
    plt.clf()
    plot_healpix_boundaries(1)
    plt.savefig('hp1.png')

    plt.clf()
    plot_healpix_boundaries(2)
    plt.savefig('hp2.png')
    



# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from numpy import *

# Writes a numpy array as PGM to the given file handle.  The numpy
# array will be converted to integer values.  If maxval=255, 8-bit
# ints; if maxval>255, 16-bit ints.
def write_pgm(x, f, maxval=255):
    (h,w) = x.shape
    if maxval >= 65536:
        raise RuntimeError('write_pgm: maxval must be < 65536')

    f.write('P5 %i %i %i\n' % (w, h, maxval))
    if maxval <= 255:
        f.write(getbuffer(x.astype(uint8)))
    else:
        f.write(getbuffer(x.astype(uint16)))


# Writes a numpy array as PPM to the given file handle.  The numpy
# array will be converted to integer values.  If maxval=255, 8-bit
# ints; if maxval>255, 16-bit ints.
def write_ppm(x, f, maxval=255):
    (h,w,d) = x.shape
    if maxval >= 65536:
        raise RuntimeError('write_pgm: maxval must be < 65536')
    if d != 3:
        raise Exception('array must have shape H x W x 3; got ' + x.shape)

    f.write('P6 %i %i %i\n' % (w, h, maxval))
    if maxval <= 255:
        f.write(getbuffer(x.astype(uint8)))
    else:
        f.write(getbuffer(x.astype(uint16)))



#! /usr/bin/env python3
from __future__ import print_function

from astrometry.util.run_command import run_command
from astrometry.util.sdss_filenames import *
from astrometry.sdss import *

def get_urls(urls, outfn, curl=False):
    for url in urls:
        if curl:
            cmd = 'curl '
            if outfn:
                cmd += '-o %s ' % outfn
            else:
                cmd += '-O '
        else:
            cmd = 'wget --continue -nv '
            if outfn:
                cmd += '-O %s ' % outfn

        cmd += '\"%s\"' % url
        print('Running:', cmd)
        (rtn, out, err) = run_command(cmd)
        if rtn == 0:
            return True
        if rtn:
            print('Command failed: command', cmd)
            print('Output:', out)
            print('Error:', err)
            print('Return val:', rtn)
    return False

def sdss_das_get_suffix(filetype):
    return ({'fpC': '.gz'}).get(filetype, '')

def sdss_das_get_url(filetype, run, camcol, field, rerun, band, suffix=None):
    if suffix is None:
        suffix = sdss_das_get_suffix(filetype)
    path = sdss_path(filetype, run, camcol, field, band, rerun)
    if path is None:
        print('Unknown SDSS filetype', filetype)
        return None
    return 'http://das.sdss.org/imaging/' + path + suffix

def sdss_das_get(filetype, outfn, run, camcol, field, band=None, reruns=None, suffix=None, gunzip=True, curl=False, ):
    if reruns is None:
        reruns = [40,41,42,44]
    urls = []
    for rerun in reruns:
        url = sdss_das_get_url(filetype, run, camcol, field, rerun, band,
                               suffix=suffix)
        if url is None:
            return False
        urls.append(url)
        
    if suffix is None:
        suffix = sdss_das_get_suffix(filetype)
    if outfn:
        outfn = outfn % { 'run':run, 'camcol':camcol, 'field':field, 'band':band } + suffix
    else:
        outfn = sdss_filename(filetype, run, camcol, field, band) + suffix
    if not get_urls(urls, outfn, curl):
        return False

    if suffix == '.gz' and gunzip:
        print('gzipped file; outfn=', outfn)
        gzipfn = outfn
        outfn = gzipfn.replace('.gz', '')
        if os.path.exists(gzipfn):
            cmd = 'gunzip -cd %s > %s' % (gzipfn, outfn)
            print('Running:', cmd)
            (rtn, out, err) = run_command(cmd)
            if rtn:
                print('Command failed: command', cmd)
                print('Output:', out)
                print('Error:', err)
                print('Return val:', rtn)
    return outfn

def sdss_das_get_fpc(run, camcol, field, band, outfn=None, reruns=None):
    return sdss_das_get('fpC', outfn, run, camcol, field, band, reruns, suffix='.gz')

def sdss_das_get_mask(run, camcol, field, band, outfn=None, reruns=None):
    return sdss_das_get('fpM', outfn, run, camcol, field, band, reruns)


if __name__ == '__main__':
    from optparse import OptionParser
    import sys

    parser = OptionParser(usage=('%prog <options> <file types>\n\n' +
                                 'file types include: fpC, fpM, fpObjc, psField, tsObj, tsField\n'
                                 'and for DR8 and above: frame'))

    parser.add_option('-r', '--run', dest='run', type='int')
    parser.add_option('-f', '--field', dest='field', type='int')
    parser.add_option('-c', '--camcol', dest='camcol', type='int')
    parser.add_option('-R', '--rerun', dest='rerun', type='int')
    parser.add_option('-b', '--band', dest='band')
    parser.add_option('-C', '--curl', dest='curl', action='store_true', default=False)
    parser.add_option('--dr9', dest='dr9', action='store_true',
                      help='Grab DR9 data')
    
    parser.set_defaults(run=None, field=None, camcol=None, band=None, rerun=None)
    (opt, args) = parser.parse_args()
    if not len(args):
        parser.print_help()
        print()
        print('Must specify types of data desired')
        sys.exit(-1)

    run = opt.run
    field = opt.field
    camcol = opt.camcol
    band = opt.band

    if run is None or field is None or camcol is None:
        parser.print_help()
        print()
        print('Must supply --run, --field, --camcol')
        sys.exit(-1)

    argdict = {}
    if band is not None:
        argdict['band'] = band
    if opt.rerun is not None:
        argdict['reruns'] = [opt.rerun]

    if opt.dr9:
        sdss = DR9()
    else:
        sdss = DR7()
        
    for filetype in args:
        print('Retrieving', filetype, '...')
        fn = sdss.retrieve(filetype, run, camcol, field, **argdict)
        #sdss_das_get(filetype, None, run, camcol, field, curl=opt.curl, **argdict)




# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE

# Creates brightstars.fits and brightstars-data.c
# from the HIP catalog from Vizier
# and the Stellarium skycultures.

# Stellarium is licensed under GPL v2+.

# I fetched hip.fits by visiting
# http://vizier.u-strasbg.fr/viz-bin/VizieR-3?-source=I/239/hip_main
# and selecting FITS binary output, and columns HIP, Vmag, _RA.icrs, _DE.icrs
# the latter of which are ICRS, epoch 2000 coordinates.

def main():
    from astrometry.util.fits import fits_table
    import numpy as np
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('skyculture', help='Stellarium skyculture/XXX/star-names.fab file')
    parser.add_argument('hip', help='Hipparcos catalog')
    parser.add_argument('out', help='Output filename')
    parser.add_argument('--brightstars', help='brightstars-data.c output filename')
    parser.add_argument('--multi', default=False, action='store_true', help='Keep multiple names (default is keep the first one only')
    parser.add_argument('--letternames', type=str, help='Stellarium stars/default/name.fab data file')
    opt = parser.parse_args()

    starnames = opt.skyculture
    hipfn = opt.hip

    H = fits_table(hipfn)
    H.ra  = H.get('_ra_icrs')
    H.dec = H.get('_de_icrs')

    hipmap = dict([(h,i) for i,h in enumerate(H.hip)])

    got_hipnums = {}
    bright = fits_table()
    bright.hip = []
    bright.ra = []
    bright.dec = []
    bright.vmag = []
    bright.names = []
    bright.letternames = []

    filenames = [(starnames, False)]
    if opt.letternames:
        filenames.append((opt.letternames, True))

    for fn,is_letter in filenames:
        f = open(fn, 'r')
        for line in f.readlines():
            line = line.strip()
            if line.startswith('#'):
                continue
            if len(line) == 0:
                continue
            hipnum = line.split('|')[0]
            hipnum = int(hipnum)
            if not is_letter:
                startstr = '_("'
                i0 = line.index(startstr) + len(startstr)
                endstr = '")'
                i1 = line.index(endstr)
                name = line[i0:i1]
            else:
                name = line.split('|')[1]
                name = name.replace('_', ' ')
            name = name.encode('ascii', 'backslashreplace')
            hipi = hipmap[hipnum]
            print('Name', name, 'HIP', hipnum, H.ra[hipi], H.dec[hipi])
            ra = H.ra[hipi]
            if not np.isfinite(ra):
                print('Null HIP entry?')
                continue

            if hipnum in got_hipnums:
                i = got_hipnums[hipnum]
                print('Appending to previous: HIP', bright.hip[i], 'names',
                      bright.letternames[i], 'and', bright.names[i])
                if is_letter:
                    bright.letternames[i].append(name)
                else:
                    bright.names[i].append(name)
                continue

            bright.hip.append(hipnum)
            bright.ra.append(H.ra[hipi])
            bright.dec.append(H.dec[hipi])
            bright.vmag.append(H.vmag[hipi])
            if is_letter:
                bright.letternames.append([name])
                bright.names.append([])
            else:
                bright.letternames.append([])
                bright.names.append([name])

            i = len(bright.hip)-1
            assert(hipnum not in got_hipnums)
            assert(bright.hip[i] == hipnum)
            got_hipnums[hipnum] = i

    if opt.multi:
        bright.names = [b' / '.join(n) for n in bright.names]
        bright.letternames = [b' / '.join(n) for n in bright.letternames]
    else:
        bright.names = [n[0] if len(n) else b'' for n in bright.names]
        bright.letternames = [n[0] if len(n) else b'' for n in bright.letternames]

    bright.to_np_arrays()
    bright.rename('letternames', 'name1')
    bright.rename('names', 'name2')
    bright.writeto(opt.out)

    if opt.brightstars:
        f = open(opt.brightstars, 'wb')
        f.write(b'// Autogenerated by brightstars.py\n')
        f.write(b'{')
        for b in bright:
            f.write((b'{ "%s", "%s", %f, %f, %f},\n' % (b.name1, b.name2, b.ra, b.dec, b.vmag)))
        f.write(b'}')
        f.close()


if __name__ == '__main__':
    main()



# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from pylab import *
from numpy import *
from math import *

# I can never remember where all the bands are...

rawdata = [
    #              min  max  mid fwhm   label-offset
    ('galex',   'fuv', 135, 175, 0, 0,  0),
    ('galex',   'nuv', 175, 280, 0, 0,  0),
    ('usnob', 'B [OJ]', 350, 540, 0, 0,      0),
    ('usnob', 'R [EF]', 590, 690, 0, 0,      1),
    ('usnob', 'N', 715, 900, 0, 0,      0),
    # Wikipedia:UBV... says: 364 nm for U, 442 nm for B, 540 nm for V.
    ('Johnson', 'U', 0,0, 364, 200, 0),
    ('Johnson', 'B', 375,500, 0,0,1),
    ('Johnson', 'V', 500,600, 0,0,0),
    ('Johnson', 'R', 550,725, 0,0,1),
    ('Johnson', 'I', 725,875, 0,0,0),
    ('sdss',  'u',   0,   0, 354,  57,  0),
    ('sdss',  'g',   0,   0, 477, 137,  1),
    ('sdss',  'r',   0,   0, 623, 137,  0),
    ('sdss',  'i',   0,   0, 763, 153,  1),
    ('sdss',  'z',   0,   0, 913,  95,  0),

    # I got these from eyeballing the synphot throughputs
    ('hst', '275', 250, 300, 0, 0,    0),
    ('hst', '336', 310, 370, 0, 0,    1),
    ('hst', '475', 400, 550, 0, 0,    0),
    ('hst', '814', 700, 950, 0, 0,    0),
    ('hst', '110', 900,1400, 0, 0,    1),
    ('hst', '160',1400,1700, 0, 0,    0),
    
    ('ukidss','Z', 830,  925, 0,   0,   0),
    ('ukidss','Y', 970, 1070, 0,   0,   1),
    ('ukidss','J',1170, 1330, 0,   0,   0),
    ('ukidss','H',1490, 1780, 0,   0,   1),
    ('ukidss','K',2030, 2370, 0,   0,   0),
    ('2mass', 'J',   0,   0, 1240, 0,   0),
    ('2mass', 'H',   0,   0, 1600, 0,   0),
    ('2mass', 'Ks',  0,   0, 2160, 0,   0),
    ('spitzer', '1', 0,   0, 3600, 0,   0),
    ('spitzer', '2', 0,   0, 4500, 0,   0),
    ('spitzer', '3', 0,   0, 5800, 0,   0),
    ('spitzer', '4', 0,   0, 8000, 0,   0),
    ]

surveys = array([survey for (survey, band, a,b,c,d,e) in rawdata])
los = array([lo or mid - fwhm/2. for (survey, band, lo, hi, mid, fwhm, e) in rawdata])
his = array([hi or mid + fwhm/2. for (survey, band, lo, hi, mid, fwhm, e) in rawdata])

survs = set(surveys)
surveymids = {}
surveylos = {}
for s in survs:
    surveymids[s] = sqrt(los[surveys == s].min() * his[surveys == s].max())
    surveylos[s] = los[surveys == s].min()

ymargin = 0.1
xmargin = 1.1

clf()
lastsurvey = ''
h = 0
for (survey, band, mn, mx, mid, fwhm, off) in rawdata:
    lo = mn or mid - fwhm/2.
    hi = mx or mid + fwhm/2.

    if survey != lastsurvey:
        h += 1
        text(surveylos[survey] / xmargin, h, survey,
             horizontalalignment='right',
             verticalalignment='center')
        lastsurvey = survey
    semilogx([lo, hi], [h, h], 'ko-')
    text(sqrt(lo*hi), h+ymargin, band,
         horizontalalignment='center',
         verticalalignment='bottom',
         fontsize='10')

    if mn:
        txt = '%g-%g' % (mn, mx)
    else:
        if fwhm:
            txt = '%g(%g)' % (mid, fwhm)
        else:
            txt = '%g' % (mid)
    text(sqrt(lo*hi), h-1.5*(ymargin*(1+off)), txt,
         horizontalalignment='center',
         verticalalignment='top',
         fontsize='7')

a=axis()
axis([70, 1.1e4, 0, h+1])

y0 = 0.1
y1 = 0.25
mx = 1.0e4
mn = 75
for (x, y, txt) in [(1e-11, y0, 'Gamma'),
                    (1e-9,  y1, 'X-ray'),
                    (400e-9,   y0, 'purple'),
                    (500e-9,   y1, 'green'),
                    (700e-9,   y0, 'red'),
                    (10e-6,   y0, 'infrared'),
                    ]:
    ha = 'left'
    x /= 1e-9
    if x >= mx:
        x = mx
        ha = 'right'
    if x < mn:
        x = mn
    text(x, y, txt,
         verticalalignment='bottom',
         horizontalalignment=ha,
         fontsize='8')

xticks([100, 1000, 10000], ['100 nm', '1 um', '10 um'])
yticks([],[])
xlabel('Wavelength')
title('Bands')
savefig('bands.png')
    

# USNOB
# http://www.iop.org/EJ/article/1538-3881/125/2/984/202452.tb1.html
# (B)
# O: 350-500 nm
# J: 385-540 nm
#  : 395-540
# (R)
# E: 620-670 nm
# F: 610-690 nm
#  : 630-690
#  : 590-690
# (N)
# N: 730-900 nm
#  : 715-900

# 2MASS
# http://www.ipac.caltech.edu/2mass/releases/allsky/doc/sec1_1.new.html
# J: 1.24 um
# H: 1.66 um
# Ks: 2.16 um

# SDSS
# http://www.astro.princeton.edu/PBOOK/camera/camera.htm
# 5 channels
# u: 354 +- 57/2 nm
# g: blue-green    477 +- 137/2 nm
# r: red           623 +- 137/2 nm
# i: far-red       763 +- 153/2 nm
# z: near-infrared 913 +- 95/2 nm

# Spitzer - IRAC
# http://www.spitzer.caltech.edu/technology/irac.shtml
# 4 channels
# 5.12 x 5.12 arcmin
# 3.6, 4.5, 5.8, and 8 microns
# 256 x 256 pixels

# GALEX
# http://www.galex.caltech.edu/QUICKFACTS/quickfacts.html
# fuv: 135 to 175 nm
# nuv: 175 to 280 nm

# HST
# many!
# http://www.stsci.edu/hst/acs/documents/handbooks/cycle17/c05_imaging2.html#356971




# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from .fits import text_table_fields

def gator2fits(fn):
    f = open(fn, 'rb')
    while True:
        line = f.readline()
        #print 'Read:', line
        if not line.startswith('\\'):
            break
    names = line
    types = f.readline()
    units = f.readline()
    xxx = f.readline()

    #print 'names:', names
    #print 'types:', types
    #print 'xxx:', xxx

    hdr = names.replace('|', ' ')

    #print 'hdr:', hdr
    tmap = {'double':float, 'int':int, 'char':str}
    types = [tmap.get(t) for t in types.replace('|',' ').split()]
    #print 'Types:', types
        
    data = f.read()
    #print 'Data:', data
    data = data.replace('|', ' ')
    data = data.replace('null', 'NaN')
    
    T = text_table_fields(None, text=data, headerline=hdr, coltypes=types)
    #T.about()
    #print 'Got', T
    return T



# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
import matplotlib.cm
import matplotlib.colors
import matplotlib.patches
import numpy as np
import pylab as plt
from matplotlib.ticker import FixedFormatter

import functools

class NanColormap(matplotlib.colors.Colormap):
    '''
    A Colormap that wraps another colormap, but replaces non-finite values
    with a fixed color.
    '''
    def __init__(self, cmap, nancolor):
        self.cmap = cmap
        self.nanrgba = matplotlib.colors.colorConverter.to_rgba(nancolor)
    def __call__(self, data, **kwargs):
        rgba = self.cmap(data, **kwargs)
        # 'data' is a MaskedArray, apparently...
        if np.all(data.mask == False):
            return rgba
        iy,ix = np.nonzero(data.mask)
        #print 'NanColormap: replacing', len(iy), 'pixels with', self.nanrgba
        # nanrgba are floats in [0,1]; convert to uint8 in [0,255].
        rgba[iy,ix, :] = np.clip(255. * np.array(self.nanrgba), 0, 255).astype(np.uint8)
        return rgba

    def __getattr__(self, name):
        ''' delegate to underlying colormap. '''
        return getattr(self.cmap, name)

def _imshow_better_defaults(imshowfunc, X, interpolation='nearest', origin='lower',
                            cmap='gray', ticks=True, **kwargs):
    '''
    An "imshow" wrapper that uses more sensible defaults.
    '''
    X = imshowfunc(X, interpolation=interpolation, origin=origin, cmap=cmap, **kwargs)
    if not ticks:
        plt.xticks([]); plt.yticks([])
    return X

def _imshow_nan(imshowfunc, X, nancolor='0.5', cmap=None, vmin=None, vmax=None, **kwargs):
    '''
    An "imshow" work-alike that replaces non-finite values by a fixed color.
    '''
    if np.all(np.isfinite(X)):
        return imshowfunc(X, cmap=cmap, vmin=vmin, vmax=vmax, **kwargs)

    # X has non-finite values.  Time to get tricky.
    cmap = matplotlib.cm.get_cmap(cmap)
    cmap = NanColormap(cmap, nancolor)
    if vmin is None or vmax is None:
        I = np.flatnonzero(np.isfinite(X))
        if vmin is None:
            try:
                vmin = X.flat[I].min()
            except ValueError:
                vmin = 0.
        if vmax is None:
            try:
                vmax = X.flat[I].max()
            except ValueError:
                vmax = 0.
    return imshowfunc(X, cmap=cmap, vmin=vmin, vmax=vmax, **kwargs)

'''
A plt.imshow() work-alike, except with defaults: interpolation='nearest', origin='lower'.
'''
imshow_better_defaults = functools.partial(_imshow_better_defaults, plt.imshow)

'''
A plt.imshow() work-alike, except handles non-finite values.  Accepts
an additional kwarg: nancolor='0.5'
'''
imshow_nan             = functools.partial(_imshow_nan, plt.imshow)

'''
My version of plt.imshow that uses imshow_better_defaults and imshow_nan.
'''
dimshow = functools.partial(_imshow_better_defaults, imshow_nan)

def replace_matplotlib_functions():
    '''
    Replaces plt.imshow with a function that handles non-finite values
    and has the defaults interpolation='nearest', origin='lower'.
    '''
    plt.imshow = dimshow




class PlotSequence(object):
    def __init__(self, basefn, format='%02i', suffix='png',
                 suffixes=None):
        self.ploti = 0
        self.basefn = basefn
        self.format = format
        if suffixes is None:
            self.suffixes = [suffix]
        else:
            self.suffixes = suffixes
        self.pattern = self.basefn + '-%s.%s'
        self.printfn = True

    def skip(self, n=1):
        self.ploti += n
    def skipto(self, n):
        self.ploti = n

    def getnextlist(self):
        #lst = ['%s-%s.%s' % (self.basefn, self.format % self.ploti, suff)
        lst = [self.pattern % (self.format % self.ploti, suff)
               for suff in self.suffixes]
        self.ploti += 1
        return lst

    def getnext(self):
        lst = self.getnextlist()
        if len(lst) == 1:
            return lst[0]
        return lst

    def savefig(self, **kwargs):
        import pylab as plt
        for fn in self.getnextlist():
            plt.savefig(fn, **kwargs)
            if self.printfn:
                print('saved', fn)

def loghist(x, y, nbins=100,
            hot=True, doclf=True, docolorbar=True, lo=0.3,
            imshowargs={},
            clampxlo=False, clampxlo_val=None, clampxlo_to=None,
            clampxhi=False, clampxhi_val=None, clampxhi_to=None,
            clampylo=False, clampylo_val=None, clampylo_to=None,
            clampyhi=False, clampyhi_val=None, clampyhi_to=None,
            clamp=None, clamp_to=None,
            **kwargs):
    #np.seterr(all='warn')
    if doclf:
        plt.clf()
    myargs = kwargs.copy()
    if not 'bins' in myargs:
        myargs['bins'] = nbins

    rng = kwargs.get('range', None)
    x = np.array(x)
    y = np.array(y)

    if not (np.all(np.isfinite(x)) and np.all(np.isfinite(y))):
        K = np.flatnonzero(np.isfinite(x) * np.isfinite(y))
        print('loghist: cutting to', len(K), 'of', len(x), 'finite values')
        x = x[K]
        y = y[K]

    if clamp is True:
        clamp = rng
    if clamp is not None:
        ((clampxlo_val, clampxhi_val),(clampylo_val, clampyhi_val)) = clamp
    if clamp_to is not None:
        ((clampxlo_to, clampxhi_to),(clampylo_to, clampyhi_to)) = clamp_to
    if clampxlo:
        if clampxlo_val is None:
            if rng is None:
                raise RuntimeError('clampxlo, but no clampxlo_val or range')
            clampxlo_val = rng[0][0]
    if clampxlo_val is not None:
        if clampxlo_to is None:
            clampxlo_to = clampxlo_val
        x[x < clampxlo_val] = clampxlo_to
    if clampxhi:
        if clampxhi_val is None:
            if rng is None:
                raise RuntimeError('clampxhi, but no clampxhi_val or range')
            clampxhi_val = rng[0][1]
    if clampxhi_val is not None:
        if clampxhi_to is None:
            clampxhi_to = clampxhi_val
        x[x > clampxhi_val] = clampxhi_to
    if clampylo:
        if clampylo_val is None:
            if rng is None:
                raise RuntimeError('clampylo, but no clampylo_val or range')
            clampylo_val = rng[1][0]
    if clampylo_val is not None:
        if clampylo_to is None:
            clampylo_to = clampylo_val
        y[y < clampylo_val] = clampylo_to
    if clampyhi:
        if clampyhi_val is None:
            if rng is None:
                raise RuntimeError('clampyhi, but no clampyhi_val or range')
            clampyhi_val = rng[1][1]
    if clampyhi_val is not None:
        if clampyhi_to is None:
            clampyhi_to = clampyhi_val
        y[y > clampyhi_val] = clampyhi_to




    (H,xe,ye) = np.histogram2d(x, y, **myargs)

    L = np.log10(np.maximum(lo, H.T))
    myargs = dict(extent=(min(xe), max(xe), min(ye), max(ye)),
                  aspect='auto',
                  interpolation='nearest', origin='lower')
    myargs.update(imshowargs)
    plt.imshow(L, **myargs)
    if hot:
        plt.hot()
    if docolorbar:
        r = [np.log10(lo)] + list(range(int(np.ceil(L.max()))))
        # print 'loghist: L max', L.max(), 'r', r
        plt.colorbar(ticks=r, format=FixedFormatter(
            ['0'] + ['%i'%(10**ri) for ri in r[1:]]))
    #set_fp_err()
    return H, xe, ye

def plothist(x, y, nbins=100, log=False,
             doclf=True, docolorbar=True, dohot=True,
             plo=None, phi=None,
             scale=None,
             imshowargs={}, **hist2dargs):
    if log:
        return loghist(x, y, nbins=nbins, doclf=doclf, docolorbar=docolorbar,
                       dohot=dohot, imshowargs=imshowargs) #, **kwargs)

    if doclf:
        plt.clf()
    (H,xe,ye) = np.histogram2d(x, y, nbins, **hist2dargs)
    if scale is not None:
        H *= scale
    myargs = dict(extent=(min(xe), max(xe), min(ye), max(ye)),
                  aspect='auto',
                  interpolation='nearest', origin='lower')
    vmin = None
    if plo is not None:
        vmin = np.percentile(H.ravel(), plo)
        myargs.update(vmin=vmin)
    if phi is not None:
        vmin = imshowargs.get('vmin', vmin)
        vmax = np.percentile(H.ravel(), phi)
        if vmax != vmin:
            myargs.update(vmax=vmax)
    myargs.update(imshowargs)
    plt.imshow(H.T, **myargs)
    if dohot:
        plt.hot()
    if docolorbar:
        plt.colorbar()
    return H, xe, ye

def setRadecAxes(ramin, ramax, decmin, decmax):
    rl,rh = ramin,ramax
    dl,dh = decmin,decmax
    rascale = np.cos(np.deg2rad((dl+dh)/2.))
    ax = [rh,rl, dl,dh]
    plt.axis(ax)
    plt.gca().set_aspect(1./rascale, adjustable='box', anchor='C')
    plt.xlabel('RA (deg)')
    plt.ylabel('Dec (deg)')
    return ax

import matplotlib.colors as mc
class ArcsinhNormalize(mc.Normalize):
    def __init__(self, mean=None, std=None, **kwargs):
        self.mean = mean
        self.std = std
        mc.Normalize.__init__(self, **kwargs)

    def _map(self, X, out=None):
        Y = (X - self.mean) / self.std
        args = (Y,)
        if out is not None:
            args = args + (out,)
        return np.arcsinh(*args)

    def __call__(self, value, clip=None):
        # copied from Normalize since it's not easy to subclass
        if clip is None:
            clip = self.clip
        result, is_scalar = self.process_value(value)
        self.autoscale_None(result)
        vmin, vmax = self.vmin, self.vmax
        if vmin > vmax:
            raise ValueError("minvalue must be less than or equal to maxvalue")
        elif vmin == vmax:
            result.fill(0)   # Or should it be all masked?  Or 0.5?
        else:
            vmin = float(vmin)
            vmax = float(vmax)
            if clip:
                mask = ma.getmask(result)
                result = ma.array(np.clip(result.filled(vmax), vmin, vmax), mask=mask)
            # ma division is very slow; we can take a shortcut
            resdat = result.data
            self._map(resdat, resdat)
            vmin = self._map(vmin)
            vmax = self._map(vmax)
            resdat -= vmin
            resdat /= (vmax - vmin)
            result = np.ma.array(resdat, mask=result.mask, copy=False)
        if is_scalar:
            result = result[0]
        return result


from matplotlib.colors import LinearSegmentedColormap

# a colormap that goes from white to black: the opposite of matplotlib.gray()
antigray = LinearSegmentedColormap('antigray',
                                   {'red':   ((0., 1, 1), (1., 0, 0)),
                                    'green': ((0., 1, 1), (1., 0, 0)),
                                    'blue':  ((0., 1, 1), (1., 0, 0))})

bluegrayred = LinearSegmentedColormap('bluegrayred',
                                      {'red':   ((0., -1, 0),
                                                 (1., 1, -1)),
                                       'green': ((0., -1,   0),
                                                 (0.5,0.5, 0.5),
                                                 (1., 0, -1)),
                                       'blue':  ((0., -1, 1),
                                                 (1., 0, -1))})

# x, y0, y1
_redgreen_data =  {'red':   ((0.,  -100,  1),
                             #(0.5,  0,  0),
                             #(0.5,  0.1, 0),
                             (0.49, 0.1, 0),
                             (0.491, 0, 0),
                             (0.51,  0, 0),
                             (0.511,  0, 0.1),
                             (1.,   0, -100)),
                   'green': ((0.,  -100,  0),
                             #(0.5,  0,  0),
                             #(0.5,  0,  0.1),
                             (0.49, 0.1, 0),
                             (0.491, 0, 0),
                             (0.51,  0, 0),
                             (0.511,  0, 0.1),
                             (1.,   1, -100)),
                   'blue':  ((0.,  -100,  0),
                             (1.,   0, -100))}
redgreen = LinearSegmentedColormap('redgreen',   _redgreen_data)

def hist_ints(x, step=1, **kwargs):
    '''
    Creates a histogram of integers.  The number of bins is set to the
    range of the data (+1).  That is, each integer gets its own bin.
    '''
    kwargs['bins'] = x.max()/step - x.min()/step + 1
    kwargs['range'] = ( (x.min()/int(step))*step - 0.5,
                        ((x.max()/int(step))*step + 0.5) )
    return plt.hist(x, **kwargs)

def hist2d_with_outliers(x, y, xbins, ybins, nout):
    '''
    Creates a 2D histogram from the given data, and returns a list of
    the indices in the data of points that lie in low-occupancy cells
    (where the histogram counts is < "nout").

    The "xbins" and "ybins" arguments are passed to numpy.histogram2d.

    You probably want to show the histogram with:

      (H, outliers, xe, ye) = hist2d_with_outliers(x, y, 10, 10, 10)
      imshow(H, extent=(min(xe), max(xe), min(ye), max(ye)), aspect='auto')
      plot(x[outliers], y[outliers], 'r.')

    Returns: (H, outliers, xe, ye)

      H: 2D histogram image
      outliers: array of integer indices of the outliers
      xe: x edges chosen by histgram2d
      ye: y edges chosen by histgram2d

    '''
    # returns (density image, indices of outliers)
    (H,xe,ye) = plt.histogram2d(x, y, (xbins,ybins))
    Out = np.array([]).astype(int)
    for i in range(len(xe)-1):
        for j in range(len(ye)-1):
            if H[i,j] > nout:
                continue
            if H[i,j] == 0:
                continue
            H[i,j] = 0
            Out = np.append(Out, np.flatnonzero((x >= xe[i]) *
                                                (x <  xe[i+1]) *
                                                (y >= ye[j]) *
                                                (y <  ye[j+1])))
    return (H.T, Out, xe, ye)


# You probably want to set the keyword radius=R
def circle(xy=None, x=None, y=None, **kwargs):
    if xy is None:
        if x is None or y is None:
            raise RuntimeError('circle: need x and y')
        xy = np.array([x,y])
    c = matplotlib.patches.Circle(xy=xy, **kwargs)
    a=plt.gca()
    c.set_clip_box(a.bbox)
    a.add_artist(c)
    return c

def ellipse(xy=None, x=None, y=None, **kwargs):
    if xy is None:
        if x is None or y is None:
            raise RuntimeError('ellipse: need x and y')
        xy = np.array([x,y])
    c = matplotlib.patches.Ellipse(xy=xy, **kwargs)
    a=plt.gca()
    c.set_clip_box(a.bbox)
    a.add_artist(c)
    return c

# return (pixel width, pixel height) of the axes area.
def get_axes_pixel_size():
    dpi = plt.gcf().get_dpi()
    figsize = plt.gcf().get_size_inches()
    axpos = plt.gca().get_position()
    pixw = figsize[0] * dpi * axpos.width
    pixh = figsize[1] * dpi * axpos.height
    return (pixw, pixh)

    # test:
    if False:
        figure(dpi=100)
        (w,h) = get_axes_pixel_size()
        # not clear why this is required...
        w += 1
        h += 1
        img = zeros((h,w))
        img[:,::2] = 1.
        img[::2,:] = 1.
        imshow(img, extent=(0,w,0,h), aspect='auto', cmap=antigray)
        xlim(0,w)
        ylim(0,h)
        savefig('imtest.png')
        sys.exit(0)

# returns (x data units per pixel, y data units per pixel)
# given the current plot range, figure size, and axes position.
def get_pixel_scales():
    a = plt.axis()
    (pixw, pixh) = get_axes_pixel_size()
    return ((a[1]-a[0])/float(pixw), (a[3]-a[2])/float(pixh))

def set_image_color_percentiles(image, plo, phi):
    # hackery...
    I = image.copy().ravel()
    I.sort()
    N = len(I)
    mn = I[max(0, int(round(plo * N / 100.)))]
    mx = I[min(N-1, int(round(phi * N / 100.)))]
    plt.gci().set_clim(mn, mx)
    return (mn,mx)




if __name__ == '__main__':

    X = np.arange(25.).reshape((5,5))
    X[2:4,3:4] = np.nan
    print(X)

    plt.clf()
    imshow_nan(X, interpolation='nearest')
    plt.savefig('1.png')

    dimshow(X)
    plt.savefig('2.png')

    replace_matplotlib_functions()

    plt.clf()
    plt.imshow(X, interpolation='nearest')
    plt.savefig('3.png')




# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
import os
import os.path

def find_data_file(fn):
    searched = []
    datadir = os.environ.get('ASTROMETRY_DATA')
    if datadir:
        pth = os.path.join(datadir, fn)
        if os.path.exists(pth):
            return pth
        searched.append(pth)
    # Add smartness here...
    dirnm = os.path.dirname(__file__)
    for i in range(5):
        pth = os.path.join(dirnm, 'data', fn)
        if os.path.exists(pth):
            return pth
        searched.append(pth)
        dirnm = os.path.dirname(dirnm)
    #pth = os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(__file__)))), 'data', fn)
    #print 'path', pth
    #if os.path.exists(pth):
    #        return pth
    print('Failed to find data file:', fn)
    print('  searched paths:', searched)
    return None




#! /usr/bin/env python3
# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE

import re
import sys
import socket
import time
from urllib.request import urlopen
from urllib.error import URLError
from urllib.parse import urlencode
from urllib.parse import urlparse, urljoin
from os.path import basename
from optparse import OptionParser
from xml.dom import minidom

from astrometry.util.starutil_numpy import *
from astrometry.util.file import *
from astrometry.util.run_command import run_command

# args: radius in deg
# returns list of local filenames to which images were written
def get_2mass_images(ra, dec, radius=1, basefn=None, band='A'):

    formvals = {
        'type': 'at', # Atlas images (not quicklook)
        'INTERSECT': 'OVERLAPS', # Images overlapping region
        'asky': 'asky', # All-sky release
        'POS': '%g %g' % (ra,dec), # RA,Dec position
        'SIZE': '%g' % radius, # Search radius (deg)
        # scan, coadd, hem, date
        'band': band, # 'A'=All bands; alternatives: J,H,K
        }

    if basefn is None:
        #basefn = '2mass-%g-%g-' % (ra,dec)
        basefn = '2mass-'

    queryurl = 'http://irsa.ipac.caltech.edu/cgi-bin/2MASS/IM/nph-im_inv'
    print('submitting form values:')
    for k,v in list(formvals.items()):
        print('  ',k,'=',v)
    print('encoded as:')
    print('  ' + urlencode(formvals))
    print()
    print('waiting for results...')
    socket.setdefaulttimeout(300)
    f = urlopen(queryurl, urlencode(formvals))
    doc = f.read()
    write_file(doc, 'res.html')

    m = re.search(r'<base href="(.*)" />', doc)
    if not m:
        raise RuntimeError('no results page: server output written to file')
    resurl = m.group(1)
    print('result base url', resurl)

    resurl += 'found.xml'

    print('requesting result url', resurl)
    res = urlopen(resurl)
    print()
    doc = res.read()
    write_file(doc, 'res2.xml')

    xmldoc = minidom.parseString(doc)

    imgs = xmldoc.getElementsByTagName('TR')
    if len(imgs) == 0:
        print('no <TR> tags found')
        return None

    urlfns = []
    for imgtag in imgs:
        print()
        if not imgtag.hasChildNodes():
            print('<TR> tag has no child node:', imgtag)
            return None
        #print 'Image:', imgtag
        tds = imgtag.getElementsByTagName('TD')
        if not len(tds):
            print('<TR> tag has no <TD> child nodes:', imgtag)
            return None

        print('Image:',  tds[0].firstChild.data)
        print('  URL:',  tds[1].firstChild.data)
        print('  Band:', tds[11].firstChild.data)
        print('  dataset (asky):', tds[18].firstChild.data)
        print('  date (yymmdd):', tds[22].firstChild.data)
        print('  hem (n/s):', tds[23].firstChild.data)
        print('  scan:', tds[24].firstChild.data)
        print('  image num:', tds[25].firstChild.data)

        url = tds[1].firstChild.data
        band = tds[11].firstChild.data
        dataset = tds[18].firstChild.data
        date = tds[22].firstChild.data
        hem = tds[23].firstChild.data
        scan = int(tds[24].firstChild.data)
        imgnum = int(tds[25].firstChild.data)

        fn = basefn + '%s_%s_%s%s%03i%04i.fits.gz' % (band, dataset, date, hem, scan, imgnum)

        urlfns.append((url,fn))

    fns = []
    for i,(url,fn) in enumerate(urlfns):
        print()
        print('Retrieving file %i of %i' % (i+1, len(urlfns)))
        print()
        # -t: num retries
        cmd = "wget -t 1 -c '%s' -O %s" % (url, fn)
        print('Running command:', cmd)

        rtn = os.system(cmd)
        # ctrl-C caught: quit, or continue?
        if os.WIFSIGNALED(rtn):
            print('wget exited with signal', os.WTERMSIG(rtn))
            break
        if os.WIFEXITED(rtn) and os.WEXITSTATUS(rtn):
            # returned non-zero.
            print('wget exited with value', os.WEXITSTATUS(rtn))
            continue

        fns.append(fn)
    return fns


if __name__ == '__main__':
    parser = OptionParser(usage='%prog [options] <ra> <dec>')

    # See http://irsa.ipac.caltech.edu/applications/2MASS/IM/inventory.html#pos

    parser.add_option('-r', dest='radius', type='float', help='Search radius, in deg (default 1 deg)')
    parser.add_option('-b', dest='basefn', help='Base filename (default: 2mass-)')
    parser.add_option('-B', dest='band', help='Band (J, H, K); default: all three')
    parser.set_defaults(radius=1.0, band='A')

    (opt, args) = parser.parse_args()
    if len(args) != 2:
        parser.print_help()
        print()
        print('Got extra arguments:', args)
        sys.exit(-1)

    # parse RA,Dec.
    ra = float(args[0])
    dec = float(args[1])

    # ugh!
    opts = {}
    for k in ['radius', 'basefn', 'band']:
        opts[k] = getattr(opt, k)

    get_2mass_images(ra, dec, **opts)



# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from math import pi,cos,sin,radians,degrees,asin,atan2,sqrt,acos,floor
from functools import reduce

def hms2ra(h, m, s):
    return 15. * (h + (m + s/60.)/60.)

def dms2dec(sign, d, m, s):
    return sign * (d + (m + s/60.)/60.)

def tokenize_hms(s):
    s = s.strip()
    # split on whitespace
    tokens = s.split()
    # also split on colons
    tokens = reduce(list.__add__, [t.split(':') for t in tokens])
    assert(len(tokens) < 4)
    assert(len(tokens) > 0)
    h = len(tokens) >= 1 and float(tokens[0]) or 0
    m = len(tokens) >= 2 and float(tokens[1]) or 0
    s = len(tokens) >= 3 and float(tokens[2]) or 0
    return (h,m,s)

def hmsstring2hms(st):
    (h,m,s) = tokenize_hms(st)
    return (h,m,s)

def hmsstring2ra(st):
    return hms2ra(*hmsstring2hms(st))

# returns (sign, deg, minutes, second)
def dmsstring2dms(s):
    sign = 1.0
    if s[0] == '-':
        sign = -1.0
        s = s[1:]
    elif s[0] == '+':
        s = s[1:]
    (d,m,s) = tokenize_hms(s)
    return (sign, d, m, s)

def dmsstring2dec(s):
    return dms2dec(*dmsstring2dms(s))

# RA in degrees
def ra2hms(ra):
    h = ra * 24. / 360.
    hh = int(floor(h))
    m = (h - hh) * 60.
    mm = int(floor(m))
    s = (m - mm) * 60.
    return (hh, mm, s)

# Dec in degrees
# returns (sign, degrees, minutes, seconds)
def dec2dms(dec):
    sgn = (dec > 0) and 1 or -1
    d = dec * sgn
    dd = int(floor(d))
    m = (d - dd) * 60.
    mm = int(floor(m))
    s = (m - mm) * 60.
    return (sgn, d, m, s)

# RA in degrees
def ra2hmsstring(ra, separator=' '):
    (h,m,s) = ra2hms(ra)
    ss = int(floor(s))
    ds = int(round((s - ss) * 1000.0))
    return separator.join(['%0.2i' % h, '%0.2i' % m, '%0.2i.%0.3i' % (ss,ds)])

# Dec in degrees
def dec2dmsstring(dec, separator=' '):
    (sign,d,m,s) = dec2dms(dec)
    # whole number of seconds
    ss = int(floor(s))
    # milliseconds
    ds = int(round((s - ss) * 1000.0))
    # fix rounding errors...
    if ds >= 1000:
        ds -= 1000
        ss += 1
    return separator.join(['%+0.2i' % int(d*sign), '%0.2i' % m, '%0.2i.%0.3i' % (ss,ds)])

# RA, Dec in degrees
def radectoxyz(ra, dec):
    rarad = radians(ra)
    decrad = radians(dec)
    cosd = cos(decrad)
    return (cosd * cos(rarad), cosd * sin(rarad), sin(decrad))

# RA, Dec in degrees
def xyztoradec(x,y,z):
    return (degrees(xy2ra(x, y)), degrees(z2dec(z)))

def xyzarrtoradec(xyz):
    return (degrees(xy2ra(xyz[0], xyz[1])), degrees(z2dec(xyz[2])))

def rad2deg(r):    return 180.0*r/pi
def deg2rad(d):    return d*pi/180.0
def rad2arcmin(r): return 10800.0*r/pi
def arcmin2rad(a): return a*pi/10800.0
def rad2arcsec(r): return 648000.0*r/pi
def arcsec2rad(a): return a*pi/648000.0
def arcsec2deg(a): return rad2deg(arcsec2rad(a))
def radec2x(r,d):  return cos(d)*cos(r) # r,d in radians
def radec2y(r,d):  return cos(d)*sin(r) # r,d in radians
def radec2z(r,d):  return sin(d)        # r,d in radians
def z2dec(z):      return asin(z)     # result in radians
def xy2ra(x,y):
    "Convert x,y to ra in radians"
    r = atan2(y,x)
    r += 2*pi*(r<0.)
    return r

def degrees_between(ra1, dec1, ra2, dec2):
    return arcsec2deg(arcsec_between(ra1, dec1, ra2, dec2))

# RA,Decs in degrees.
def arcsec_between(ra1, dec1, ra2, dec2):
    xyz1 = radectoxyz(ra1, dec1)
    xyz2 = radectoxyz(ra2, dec2)
    d2 = sum([(a-b)**2 for (a,b) in zip(xyz1, xyz2)])
    return distsq2arcsec(d2)

def rad2distsq(rad):
    return 2. * (1. - cos(rad))

def arcsec2distsq(arcsec):
    return rad2distsq(arcsec2rad(arcsec))

def arcsec2dist(arcsec):
    return sqrt(arcsec2distsq(arcsec))

def dist2arcsec(dist):
    return distsq2arcsec(dist**2)

def distsq2arcsec(dist2):
    return rad2arcsec(distsq2rad(dist2))

def distsq2rad(dist2):
    return acos(1. - dist2 / 2.)



#! /usr/bin/env python3
# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
import os
import sys
import fnmatch
from astrometry.plot.plotstuff_c import *
from numpy import *
from optparse import OptionParser

if __name__ == '__main__':
    parser = OptionParser(usage='%prog [options] <ra> <dec> <width-in-deg> <index file> <output-filename>')
    (opt, args) = parser.parse_args()
    if len(args) != 5:
        parser.print_help()
        print()
        print('Got wrong number of arguments:', args)
        sys.exit(-1)

    ra = float(args[0])
    dec = float(args[1])
    width = float(args[2])
    indexfn = args[3]
    outfn = args[4]

    pargs = plotstuff_new()
    pargs.outformat = PLOTSTUFF_FORMAT_PNG
    pargs.outfn = outfn

    plotstuff_set_size(pargs, 800, 800)
    plotstuff_set_wcs_box(pargs, ra, dec, width)

    plotstuff_set_color(pargs, 'verydarkblue')
    plotstuff_run_command(pargs, 'fill')

    plotstuff_set_color(pargs, 'gray')
    plotstuff_set_alpha(pargs, 0.25)
    grid = plot_grid_get(pargs)
    grid.rastep = grid.decstep = 1
    grid.ralabelstep = grid.declabelstep = 0
    plotstuff_run_command(pargs, 'grid')
    plotstuff_set_color(pargs, 'gray')
    grid.rastep = grid.decstep = 0
    grid.ralabelstep = grid.declabelstep = 1
    plotstuff_set_bgrgba2(pargs, 0, 0, 0, 0.5)
    plotstuff_run_command(pargs, 'grid')

    plotstuff_set_color(pargs, 'green')
    pargs.markersize = 1
    index = plot_index_get(pargs)
    plot_index_add_file(index, indexfn)
    index.stars = 1
    index.quads = 1
    plotstuff_run_command(pargs, 'index')

    plotstuff_output(pargs)

    plotstuff_free(pargs)
    



# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
def shell_escape(s):
    repl = ('\\', '|', '&', ';', '(', ')', '<', '>', ' ', '\t',
            '\n', '$', "'", '"', "`")
    # (note, \\ must be first!)
    for x in repl:
        s = s.replace(x, '\\'+x)
    return s

# escape a string that will appear inside double-quotes.
def shell_escape_inside_quotes(s):
    repl = ('\\', '\t', '`', '"', '$')
    # (note, \\ must be first!)
    for x in repl:
        s = s.replace(x, '\\'+x)
    return s



#! /usr/bin/env python3
# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE

import sys

import http.client
http.client.HTTPConnection.debuglevel = 1 

import urllib.request, urllib.parse, urllib.error
import urllib.request, urllib.error, urllib.parse
#from xml.parsers import expat
import xml.dom
from xml.dom import minidom
import pickle
import os
import os.path
import re
import time

from numpy import *

from astrometry.sdss.sqlcl import query as casquery
from astrometry.util.file import *

class Cas(object):
    def __init__(self, **kwargs):
        self.params = kwargs

    def get_url(self, relurl):
        return self.params['base_url'] + relurl

    def login_url(self):
        return self.get_url('login.aspx')

    def submit_url(self):
        return self.get_url(self.params['submiturl'])

    def mydb_url(self):
        return self.get_url('MyDB.aspx')

    def mydb_index_url(self):
        return self.get_url('mydbindex.aspx')

    def mydb_action_url(self):
        return self.get_url(self.params['actionurl'])

    def drop_url(self):
        return self.get_url('DropTable.aspx?tableName=%s')

    def output_url(self):
        return self.get_url('Output.aspx')

    def job_details_url(self):
        return self.get_url('jobdetails.aspx?id=%i')

    def cancel_url(self):
        return self.get_url('cancelJob.aspx')

    def submit_query(self, sql, table='', taskname='', dbcontext=None):
        if dbcontext is None:
            dbcontext = self.params['defaultdb']

        # MAGIC
        data = urllib.parse.urlencode({
            'targest': dbcontext,
            'sql': sql,
            'queue': 500,
            'syntax': 'false',
            'table': table,
            'taskname': taskname,
            })
        f = urllib.request.urlopen(self.submit_url(), data)
        doc = f.read()
        redirurl = f.geturl()
        # older CasJobs version redirects to the job details page:
        # just pull the jobid out of the redirected URL.
        print('Redirected to URL', redirurl)
        pat = re.escape(self.job_details_url().replace('%i','')) +  '([0-9]*)'
        #print 'pattern:', pat
        m = re.match(pat, redirurl)
        if m is not None:
            jobid = int(m.group(1))
            print('jobid:', jobid)
            return jobid
        #write_file(doc, 'sub.out')
        xmldoc = minidom.parseString(doc)
        jobids = xmldoc.getElementsByTagName('jobid')
        if len(jobids) == 0:
            print('No <jobid> tag found:')
            print(doc)
            return None
        if len(jobids) > 1:
            print('Multiple <jobid> tags found:')
            print(doc)
            return None
        jobid = jobids[0]
        if not jobid.hasChildNodes():
            print('<jobid> tag has no child node:')
            print(doc)
            return None
        jobid = jobid.firstChild
        if jobid.nodeType != xml.dom.Node.TEXT_NODE:
            print('job id is not a text node:')
            print(doc)
            return None
        jobid = int(jobid.data)
        if jobid == -1:
            # Error: find error message.
            print('Failed to submit query.  Looking for error message...')
            founderr = False
            msgs = xmldoc.getElementsByTagName('message')
            for msg in msgs:
                if msg.hasChildNodes():
                    c = msg.firstChild
                    if c.nodeType == xml.dom.Node.TEXT_NODE:
                        print('Error message:', c.data)
                        founderr = True
            if not founderr:
                print('Error message not found.  Whole response document:')
                print()
                print(doc)
                print()
        return jobid
    
    def login(self, username, password):
        print('Logging in.')
        data = urllib.parse.urlencode({'userid': username, 'password': password})
        f = urllib.request.urlopen(self.login_url(), data)
        d = f.read()
        # headers = f.info()
        # print 'headers:', headers
        # print 'Got response:'
        # print d
        return None

    def cancel_job(self, jobid):
        data = urllib.parse.urlencode({'id': jobid, 'CancelJob': 'Cancel Job'})
        f = urllib.request.urlopen(self.cancel_url(), data)
        f.read()

    # Returns 'Finished', 'Ready', 'Started', 'Failed', 'Cancelled'
    def get_job_status(self, jobid):
        # print 'Getting job status for', jobid
        url = self.job_details_url() % jobid
        print('Job details URL:', url)
        doc = urllib.request.urlopen(url).read()
        for line in doc.split('\n'):
            for stat in ['Finished', 'Ready', 'Started', 'Failed', 'Cancelled']:
                if ('<td > <p class = "%s">%s</p></td>' %(stat,stat) in line or
                    #'<td > <p class="%s">%s</p></td>' %(stat,stat) in line or
                    '<p class="%s">%s</p>' % (stat,stat) in line or
                    '<td class="center"> <p class = "%s">%s</p></td>' %(stat,stat) in line or
                    '<td class="center"> <p class = "%s">Running</p></td>' %(stat) in line): # Galex "Started"/Running
                    return stat
        return None

    def drop_table(self, dbname):
        url = self.drop_url() % dbname
        try:
            f = urllib.request.urlopen(url)
        except Exception as e:
            print('Failed to drop table', dbname)
            print(e)
            return False
        doc = f.read()
        (vs,ev) = get_viewstate_and_eventvalidation(doc)
        data = urllib.parse.urlencode({'yesButton':'Yes',
                                 '__EVENTVALIDATION':ev,
                                 '__VIEWSTATE':vs})
        print('Dropping table', dbname)
        try:
            f = urllib.request.urlopen(url, data)
        except Exception as e:
            print('Failed to drop table', dbname)
            print(e)
            return False
        d = f.read()
        # print 'Got response:'
        # print d
        # write_file(d, 'res.html')
        return True

    def request_output(self, mydbname):
        url = self.mydb_action_url() % mydbname
        try:
            # Need to prime the VIEWSTATE by "clicking" through...
            f = urllib.request.urlopen(self.mydb_url())
            f.read()
            f = urllib.request.urlopen(self.mydb_index_url())
            f.read()
            # request = urllib2.Request(url)
            # request.add_header('User-Agent', 'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.9.0.11) Gecko/2009060214 Firefox/3.0.11')
            # f = urllib2.urlopen(request)
            # Referer: http://galex.stsci.edu/casjobs/mydbindex.aspx
            f = urllib.request.urlopen(url)
        except urllib.error.HTTPError as e:
            print('HTTPError:', e)
            print('  code', e.code)
            print('  msg', e.msg)
            print('  hdrs', e.hdrs)
            print('  data:', e.fp.read())
            raise e
        doc = f.read().strip()
        # write_file(doc, 'r1.html')
        (vs,ev) = get_viewstate_and_eventvalidation(doc)
        data = {'extractDDL':'FITS', 'Button1':'Go',
                '__VIEWSTATE':vs}
        if ev is not None:
            data['__EVENTVALIDATION'] = ev
        extra = self.params.get('request_output_extra')
        if extra is not None:
            data.update(extra)
            print('requesting FITS output of MyDB table', mydbname)
        print('url', url)
        print('data', urllib.parse.urlencode(data))
        try:
            f = urllib.request.urlopen(url, urllib.parse.urlencode(data))
        except urllib.error.HTTPError as e:
            print('HTTPError:', e)
            print('  code', e.code)
            # print '  reason', e.reason
            raise e
        d = f.read()
        # print 'Got response:'
        # print d
        # write_file(d, 'res.html')
        return

    def get_ready_outputs(self):
        url = self.output_url()
        print('Hitting output URL', url)
        f = urllib.request.urlopen(url)
        doc = f.read()
        write_file(doc, 'ready.html')
        print('Wrote ready downloads to ready.html')
        urls = []
        fns = []

        # split into successful and failed parts
        ifailed = doc.index('Failed Output:')
        if ifailed > -1:
            failed = doc[ifailed:]
            doc = doc[:ifailed]
        else:
            failed = ''
        
        rex = re.compile('<a href="(' + re.escape(self.params['outputbaseurl']) + '(.*))">Download</a>')
        for line in doc.split('\n'):
            m = rex.search(line)
            if not m:
                continue
            urls.append(m.group(1))
            fns.append(m.group(2))

        
        fails = []
        rex = re.compile(r'^\s*<td>(.*?)</td><td>.*?</td><td>(.*?)</td><td>(.*?)</td>', flags=re.MULTILINE | re.DOTALL)
        for m in rex.finditer(failed):
            tab = m.group(1)
            if '<br>' in tab:
                continue
            msg = m.group(3).replace('\r', ' ').replace('\n', ' ')
            fails.append((tab, m.group(2), msg))
            # ( table name, date string, failure message )
        return (urls, fns, fails)

    def sql_to_fits(self, sql, outfn, dbcontext=None, sleeptime=10):
        import random
        dbname = ''.join(chr(ord('A') + random.randrange(26))
                         for x in range(10))
        if sql.lower().startswith('select '):
            sql = 'select into mydb.%s' % dbname + sql[6:]
        print('Submitting query: "%s"' % sql)
        if dbcontext is not None:
            kwargs = dict(dbcontext=dbcontext)
        else:
            kwargs = {}
        jid = self.submit_query(sql, **kwargs)
        print('Submitted job id', jid)
        print('Waiting for job id:', jid)
        while True:
            jobstatus = self.get_job_status(jid)
            print('Job id', jid, 'is', jobstatus)
            if jobstatus in ['Finished', 'Failed', 'Cancelled']:
                break
            print('Sleeping...')
            time.sleep(sleeptime)
        print('Output-downloads-delete')
        dodelete = True
        self.output_and_download([dbname], [outfn], dodelete)
    
    # Requests output of the given list of databases, waits for them to appear,
    # downloads them, and writes them to the given list of local filenames.
    #
    # 'dbs' and 'fns' must be either strings,
    #   or lists of string of the same length.
    #
    # If 'dodelete' is True, the databases will be deleted after download.
    #
    def output_and_download(self, dbs, fns, dodelete=False, sleeptime=10,
                            raiseonfail=True):
        if type(dbs) is str:
            dbs = [dbs]
            assert(type(fns) is str)
            fns = [fns]
        # copy lists
        dbs = dbs[:]
        fns = fns[:]
            
        print('Getting list of available downloads...')
        (preurls,nil,prefails) = self.get_ready_outputs()
        #print 'Preurls:', preurls
        #print 'Prefails:', prefails
        for db in dbs:
            print('Requesting output of', db)
            self.request_output(db)
        while True:
            print('Waiting for output to appear...')
            (durls,dfns,dfails) = self.get_ready_outputs()
            (newurls, newfns) = find_new_outputs(durls, dfns, preurls)
            #print 'URLs:', durls
            #print 'New URLs:', newurls
            print('New outputs available:', newfns)
            newfails = [f for f in dfails if not f in prefails]
            print('New failures:', newfails)
            for (fn,db) in zip(fns,dbs):
                for (dfn,durl) in zip(newfns,newurls):
                    # the filename will contain the db name.
                    if not db in dfn:
                        continue
                    print('Output', dfn, 'looks like it belongs to database', db)
                    print('Downloading to local file', fn)
                    cmd = 'wget -O "%s" "%s"' % (fn, durl)
                    print('  (running: "%s")' % cmd)
                    w = os.system(cmd)
                    if not os.WIFEXITED(w) or os.WEXITSTATUS(w):
                        print('download failed.')
                        return -1
                    dbs.remove(db)
                    fns.remove(fn)
                    if dodelete:
                        print('Deleting database', db)
                        self.drop_table(db)
            for (fn,db) in zip(fns,dbs):
                for tab,date,msg in newfails:
                    if db != tab:
                        continue
                    print('Failure:', tab, 'date', date, 'message', msg)
                    print('looks like it belongs to database', db)
                    dbs.remove(db)
                    fns.remove(fn)
                    if raiseonfail:
                        raise RuntimeError('Error from CasJobs on output of table "%s": "%s"' % (db, msg))
                    
            if not len(dbs):
                   break
            print('Waiting...')
            time.sleep(sleeptime)
        return 0


def get_viewstate_and_eventvalidation(doc):
    rex = re.compile('<input type="hidden" name="__VIEWSTATE" (?:id="__VIEWSTATE" )?value="(.*)" />')
    vs = None
    for line in doc.split('\n'):
        m = rex.search(line)
        if not m:
            continue
        vs = m.group(1)
        break
    rex = re.compile('<input type="hidden" name="__EVENTVALIDATION" id="__EVENTVALIDATION" value="(.*)" />')
    ev = None
    for line in doc.split('\n'):
        m = rex.search(line)
        if not m:
            continue
        ev = m.group(1)
        break
    return (vs,ev)

def query(sql):
    f = casquery(sql)
    header = f.readline().strip()
    if header.startswith('ERROR'):
        raise RuntimeError('SQL error: ' + f.read())
    cols = header.split(',')
    results = []
    for line in f:
        words = line.strip().split(',')
        row = []
        for w in words:
            try:
                ival = int(w)
                row.append(ival)
                continue
            except ValueError:
                pass
            try:
                fval = float(w)
                row.append(fval)
                continue
            except ValueError:
                pass
            row.append(w)
            results.append(row)
    return (cols, results)

def setup_cookies():
    cookie_handler = urllib.request.HTTPCookieProcessor()
    opener = urllib.request.build_opener(cookie_handler)
    # ...and install it globally so it can be used with urlopen.
    urllib.request.install_opener(opener)

def find_new_outputs(durls, dfns, preurls):
    newurls = [u for u in durls if not u in preurls]
    newfns =  [f for (f,u) in zip(dfns,durls) if not u in preurls]
    return (newurls, newfns)

def get_known_servers():
    return     {
        'galex': Cas(
            base_url='http://galex.stsci.edu/casjobs/',
            submiturl='SubmitJob.aspx',
            actionurl='mydbcontent.aspx?tableName=%s&kind=tables',
            defaultdb='GALEXGR4Plus5',
            outputbaseurl='http://mastweb.stsci.edu/CasOutPut/FITS/'
            ),
        # SDSS
        'dr7': Cas(
            base_url='http://casjobs.sdss.org/CasJobs/',
            submiturl='submitjobhelper.aspx',
            actionurl='mydbcontent.aspx?ObjName=%s&ObjType=TABLE&context=MyDB&type=normal',
            defaultdb='DR7',
            request_output_extra={ 'targetDDL':'TargDR7Long' },
            outputbaseurl='http://casjobs.sdss.org/CasJobsOutput2/FITS/'
            ),
        # SDSS
        'dr8': Cas(
            # These will change...
            base_url='http://skyserver.sdss3.org/casjobs/',
# base_url='http://skyservice.pha.jhu.edu/casjobs/',
            submiturl='submitjobhelper.aspx',
            actionurl='mydbcontent.aspx?ObjName=%s&ObjType=TABLE&context=MyDB&type=normal',
            defaultdb='DR8',
            #request_output_extra={ 'targetDDL':'Thumper_DR7' },
            outputbaseurl='http://skyservice.pha.jhu.edu/CasJobsOutput/FITS/'
            ),
        'dr9': Cas(
            # These will change...
            base_url='http://skyserver.sdss3.org/casjobs/',
            submiturl='submitjobhelper.aspx',
            actionurl='mydbcontent.aspx?ObjName=%s&ObjType=TABLE&context=MyDB&type=normal',
            defaultdb='DR9',
            outputbaseurl='http://skyservice.pha.jhu.edu/CasJobsOutput/FITS/'
            ),
            }


if __name__ == '__main__':
    from optparse import OptionParser

    surveys = get_known_servers()

    parser = OptionParser(usage=('%prog <options> <args>'))
    parser.add_option('-s', '--survey', dest='survey', default='dr7',
                      help=('Set the CasJobs instance to use: one of: ' +
                            ', '.join(list(surveys.keys()))))
    parser.add_option('-c', '--context', '--db', dest='dbcontext',
                      help='Database context ("DR7", "Stripe82", etc)')
    opt,args = parser.parse_args()
    cas = surveys[opt.survey]

    if len(args) < 2:
        print('%s <username> <password> [command <args>]' % sys.argv[0])
        print()
        print('commands include:')
        print('   sqltofits (<sql> or <@sql-file>) output.fits')
        print('   delete <database> [...]')
        print('   query  ( <sql> or <@sqlfile> ) [...]')
        print('   querywait  ( <sql> or <@sqlfile> ) [...]')
        print('     -> submit query and wait for it to finish')
        print('   output <database> [...]')
        print('     -> request that a database be output as a FITS table')
        print('   outputdownload <database> <filename> [...]')
        print('     -> request output, wait for it to finish, and download to <filename>')
        print('   outputdownloaddelete <database> <filename> [...]')
        print('     -> request output, wait for it to finish, download to <filename>, and drop table.')
        sys.exit(-1)

    instance = 4

    username = args[0]
    password = args[1]
    cmd = None
    if len(args) > 2:
        cmd = args[2]

    setup_cookies()

    cas.login(username, password)

    if cmd == 'delete':
        if len(args) < 4:
            print('Usage: ... delete <db>')
            sys.exit(-1)
        db = args[3]
        print('Dropping', db)
        cas.drop_table(db)
        sys.exit(0)

    if cmd in ['query', 'querywait']:
        qs = args[3:]
        if len(qs) == 0:
            print('Usage: ... query <sql> or <@file> [...]')
            sys.exit(-1)
        jids = []
        for q in qs:
            if q.startswith('@'):
                q = read_file(q[1:])
            print('Submitting query: "%s"' % q)
            jobid = cas.submit_query(q, dbcontext=opt.dbcontext)
            print('Submitted job id', jobid)
            jids.append(jobid)
        if cmd in ['querywait']:
            # wait for them to finish.
            while True:
                print('Waiting for job ids:', jids)
                for jid in jids:
                    jobstatus = cas.get_job_status(jid)
                    print('Job id', jid, 'is', jobstatus)
                    if jobstatus in ['Finished', 'Failed', 'Cancelled']:
                        jids.remove(jid)
                if not len(jids):
                    break
                print('Sleeping...')
                time.sleep(10)
        sys.exit(0)

    if cmd == 'sqltofits':
        if len(args) != 5:
            print('Usage: ... sqltofits [<sql> or <@file>] output.fits')
            sys.exit(-1)
        q = args[3]
        outfn = args[4]
        if q.startswith('@'):
            q = read_file(q[1:])
        cas.sql_to_fits(q, outfn, dbcontext=opt.dbcontext)
        sys.exit(0)
            
    if cmd == 'output':
        dbs = args[3:]
        if len(dbs) == 0:
            print('Usage: ... output <db> [<db> ...]')
            sys.exit(-1)
        for db in dbs:
            print('Requesting output of db', db)
            cas.request_output(db)
        sys.exit(0)

    if cmd in ['outputdownload', 'outputdownloaddelete']:
        dodelete = (cmd == 'outputdownloaddelete')
        dbfns = args[3:]
        if len(dbfns) == 0 or len(dbfns) % 2 == 1:
            print('Usage: ... outputdownload <db> <filename> [...]')
            sys.exit(-1)
        dbs = dbfns[0::2]
        fns = dbfns[1::2]

        cas.output_and_download(dbs, fns, dodelete)
        sys.exit(0)

    print('Unrecognized command')
    



# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE

from math import *  # Needed if we're going to exclude pylab :)

PIl = 3.1415926535897932384626433832795029
def rad2deg(r):    return 180.0*r/PIl
def deg2rad(d):    return d*PIl/180.0
def rad2arcmin(r): return 10800.0*r/PIl
def arcmin2rad(a): return a*PIl/10800.0
def rad2arcsec(r): return 648000.0*r/PIl
def arcsec2rad(a): return a*PIl/648000.0
def radec2x(r,d):  return cos(d)*cos(r) # r,d in radians
def radec2y(r,d):  return cos(d)*sin(r) # r,d in radians
def radec2z(r,d):  return sin(d)        # r,d in radians
def z2dec(z):      return arcsin(z)     # result in radians
def xy2ra(x,y):
    "Convert x,y to ra in radians"
    r = arctan2(y,x)
    r += 2*PIl*(r>=0.00)
    return r

# wrapper to give us a tuple.
def radec2xyz(r,d): return (radec2x(r,d), radec2y(r,d), radec2z(r,d))

# Takes min/max RA & DEC (which a number of the catalogs provide) and
# compute the xyz coordinates of the box on the surface of the sphere.
def radecbox2xyz(min_r, max_r, min_d, max_d):
    return [radec2xyz(min_r, min_d), radec2xyz(min_r, max_d), \
        radec2xyz(max_r, max_d), radec2xyz(max_r, min_d)]



#define radscale2xyzscale(r) (sqrt(2.0-2.0*cos(r/2.0)))
def star_coords(s,r):
    # eta is a vector perpendicular to r
    etax = -r[1]
    etay = +r[0]
    etaz = 0.0
    eta_norm = sqrt(etax * etax + etay * etay)
    etax /= eta_norm
    etay /= eta_norm

    # xi =  r cross eta
    xix = -r[2] * etay
    xiy =  r[2] * etax
    xiz =  r[0] * etay - r[1] * etax
    sdotr = dot(s,r)

    return (
            s[0] * xix / sdotr +
            s[1] * xiy / sdotr +
            s[2] * xiz / sdotr,
            s[0] * etax / sdotr +
            s[1] * etay / sdotr
           )

def radectohealpix(ra, dec):
    x = radec2x(ra, dec)
    y = radec2y(ra, dec)
    z = radec2z(ra, dec)
    return xyztohealpix(x, y, z)

def xyztohealpix(x, y, z) :
    "Return the healpix catalog number associated with point (x,y,z)"

    # the polar pixel center is at (z,phi/pi) = (2/3, 1/4)
    # the left pixel center is at (z,phi/pi) = (0, 0)
    # the right pixel center is at (z,phi/pi) = (0, 1/2)
    twothirds = 2.0 / 3.0

    phi = arctan2(y, x)
    if phi < 0.0:
        phi += 2.0 * pi

    phioverpi = phi / pi

    if z >= twothirds or z <= -twothirds:
        # definitely in the polar regions.
        # the north polar healpixes are 0,1,2,3
        # the south polar healpixes are 8,9,10,11
        if z >= twothirds:
            offset = 0
        else:
            offset = 8

        pix = int(phioverpi * 2.0)

        return offset + pix

    # could be polar or equatorial.
    offset = int(phioverpi * 2.0)
    phimod = phioverpi - offset * 0.5

    z1 =  twothirds - (8.0 / 3.0) * phimod
    z2 = -twothirds + (8.0 / 3.0) * phimod

    if z >= z1 and z >= z2:
        # north polar
        return offset
    if z <= z1 and z <= z2:
        # south polar
        return offset + 8
    if phimod < 0.25:
        # left equatorial
        return offset + 4
    # right equatorial
    return ((offset+1)%4) + 4


# Brought over from starutil.c & plotcat.c
def project_equal_area(point):
    x, y, z = point
    xp = x * sqrt(1. / (1. + z))
    yp  = y * sqrt(1. / (1. + z))
    xp = 0.5 * (1.0 + xp)
    yp = 0.5 * (1.0 + yp)
    return (xp, yp)

def project_hammer_aitoff_x(point):
    x, y, z = point
    theta = atan(float(x) / (z + 0.000001))
    r = sqrt(x * x + z * z)
    if z < 0:
        if x < 0:
            theta = theta - pi 
        else:
            theta = pi + theta
    theta /= 2.0
    zp = r * cos(theta)
    xp = r * sin(theta)
    assert zp >= -0.01
    return project_equal_area((xp, y, zp))


def getxy(projectedpoint, N):
    px, py = projectedpoint
    px = 0.5 + (px - 0.5) * 0.99
    py = 0.5 + (py - 0.5) * 0.99
    X = round(px * 2*N)
    Y = round(py * N)
    return (X, Y)




# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE

#import sqlite3
#import commands
from __future__ import print_function
import os
import sys

def backtick(cmd):
    (childin, childout) = os.popen2(cmd)
    childin.close()
    out = childout.read()
    return out

def getval(fn, key):
    cmd = "sqlite " + fn + " \"select val from jobdata where key like '" + key + "'\""
    out = backtick(cmd)
    return out

#fn = '/home/gmaps/ontheweb-data/alpha/200705/00001872/jobdata.db'
#email = getval(fn, 'email');
#status = getval(fn, 'checked-done')
#print email, status

outdir = "/data2/apod-solves/"

for month in sys.argv[1:]:
    #print month
    jobs = os.listdir(month)
    for job in jobs:
        path = month + '/' + job + '/'
        fn = path + "jobdata.db"
        #print fn
        email = getval(fn, 'email').strip()
        #print email
        if email != 'ckochanek@astronomy.ohio-state.edu':
            continue
        solvedfn = path + 'solved'
        if not os.path.exists(solvedfn):
            continue
        #status = getval(fn, 'checked-done')
        #print job, status
        #wcsfn = path + "wcs.fits"
        #print wcsfn
        #print job
        print(path)
        cmd = "pngtopnm " + path + "fullsize.png | pnmtojpeg > " + outdir + "csk-" + job + ".jpg"
        backtick(cmd)
        cmd = "cp " + path + "wcs.fits " + outdir + "csk-" + job + ".wcs"
        backtick(cmd)


#conn = sqlite3.connect(fn)
#cur = conn.cursor()
#cur.execute("select val from jobdata where key like 'user-email'")




# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
import math
try:
    import pyfits
except ImportError:
    try:
        from astropy.io import fits as pyfits
    except ImportError:
        raise ImportError("Cannot import either pyfits or astropy.io.fits")
from numpy import *
from matplotlib.pylab import figure, plot, xlabel, ylabel, loglog, clf
from matplotlib.pylab import semilogy, show, find, legend, hist, axis

def plotDvsR(ix, iy, fx, fy, R):
    IR = argsort(R)

    mR = []
    mD = []
    mI = []

    # Look at index stars in order of R.
    for ii in range(len(IR)):
        i = IR[ii]
        x = ix[i]
        y = iy[i]
        D = sqrt((fx - x)**2 + (fy - y)**2)
        # Grab field stars within a matching radius.
        r = 5
        I = find( D < r )
        for j in I:
            mR.append(R[i])
            mD.append(D[j])
            mI.append(ii)
    plot(mR, mD, 'ro')
    xlabel('Distance from quad center')
    ylabel('Match distance')

def plotIvsR(ix, iy, fx, fy, cx, cy):
    RF = sqrt((fx - cx)**2 + (fy - cy)**2)
    mD2 = []
    mI2 = []
    for i in range(len(RF)):
        r = 5
        D = sqrt((ix - fx[i])**2 + (iy - fy[i])**2)
        I = find( D < r )
        for j in I:
            mD2.append(D[j])
            mI2.append(i)
    #plot(mI, mD, 'ro', mI2, mD2, 'bo')
    plot(mI2, mD2, 'bo')
    xlabel('Index star number')
    ylabel('Match distance')
    #legend(('Sorted by dist from quad center', 'Sorted by brightness'))


    

if __name__ == '__main__':

    # Index stars
    ixy = pyfits.open('ver/index.xy.fits')
    ixy = ixy[1].data
    ix = ixy.field(0)
    iy = ixy.field(1)
    NI = len(ix)

    # Field stars
    fxy = pyfits.open('ver/field.xy.fits')
    fxy = fxy[1].data
    fx = fxy.field(0)
    fy = fxy.field(1)
    NF = len(fx)

    # The matched quad.
    mf = pyfits.open('ver/match.fits')
    mf = mf[1].data
    quad = mf.field('quadpix')[0]
    quad = quad[0:8].reshape(4,2)
    qx = quad[:,0]
    qy = quad[:,1]
    # Quad center.
    cx = mean(qx)
    cy = mean(qy)

    # Grab index stars that are within the field.
    iok = find( (ix > min(fx)) *
                (ix < max(fx)) *
                (iy > min(fy)) *
                (iy < max(fy)) )
    ix = [ix[i] for i in iok]
    iy = [iy[i] for i in iok]


    figure(1)
    clf()
    I = [0, 2, 1, 3, 0]
    plot(
        [cx], [cy], 'ro',
        qx[I], qy[I], 'r-',
        ix, iy, 'rx',
        fx, fy, 'b+',
        )

    # RMS quad radius
    RQ = sqrt(sum((qx - cx)**2 + (qy - cy)**2) / 4)

    # Distance from quad center.
    RI = sqrt((ix - cx)**2 + (iy - cy)**2)
    RF = sqrt((fx - cx)**2 + (fy - cy)**2)

    # Angle from quad center.
    AI = array([math.atan2(y - cy, x - cx) for (x,y) in zip(ix,iy)])
    AF = array([math.atan2(y - cy, x - cx) for (x,y) in zip(fx,fy)])

    # Look at index stars in order of R.
    IR = argsort(RI)
    allD = array([])
    allDR = array([])
    allDA = array([])
    allR = array([])
    allDist = array([])

    for i in IR:
        # regularizer...
        reg = RQ
        #
        Distscale = 1
        DRscale = 1 / RQ

        dR = ((RI[i] + reg) / (RF + reg)) - 1.0
        dA =  AI[i] - AF
        # handle wrap-around
        absdA = abs(dA)
        absdA = vstack((absdA, abs(absdA - 2*math.pi))).min(axis=0)
        #D = sqrt(dR**2 + dA**2)
        D = sqrt(dR**2 + absdA**2)
        D = D / DRscale

        Dist = sqrt((ix[i] - fx)**2 + (iy[i] - fy)**2)
        Dist = Dist / Distscale

        iSmall = array(find((D < 1) + (Dist < 1)))
        allD = hstack((allD, D[iSmall]))
        allDR = hstack((allDR, dR[iSmall]))
        allDA = hstack((allDA, dA[iSmall]))
        allR = hstack((allR, repeat(RI[i], len(iSmall))))
        allDist = hstack((allDist, Dist[iSmall]))

    figure(2)
    clf()
    plot(allDR, allDA, 'ro', ms=1)
    xlabel('DR')
    ylabel('DA')

    figure(3)
    clf()
    #plot(allR/RQ, allD, 'r.')
    #plot(allR/RQ, vstack((allD, allDist)).min(axis=0), 'r.')
    plot(
        #allR/RQ, vstack((allD, allDist)).min(axis=0), 'mo',
        allR/RQ, allD, 'r+',
        allR/RQ, allDist, 'bx'
        )
    xlabel('R (quad radiuses)')
    #ylabel('min( Dist, D(R+A) )')
    ylabel('Dist, D(R,A)')
    a = axis()
    axis([a[0], a[1], 0, 2.0])
    legend(('D(R,A)', 'D'))

    #allDist = array([])
    #allRDist = array([])
    #for i in IR:
    #    Dist = sqrt((ix[i] - fx)**2 + (iy[i] - fy)**2)
    #    iSmall = array(find(Dist < 5))
    #    allDist = hstack((allDist, Dist[iSmall]))
    #    allRDist = hstack((allRDist, repeat(RI[i], len(iSmall))))
    #figure(4)
    #clf()
    #plot(allRDist/RQ, allDist, 'r.')
    #xlabel('R')
    #ylabel('Dist')

    #figure(2)
    #clf()
    #plotDvsR(ix, iy, RI, fx, fy)

    #figure(3)
    #clf()
    #plotDvsI(ix, iy, fx, fy, cx, cy)

    



from __future__ import print_function
from setuptools import setup, Extension
from setuptools.command.install import install
import subprocess
import os
import sys
from glob import glob

version = '0.0'
try:
    subprocess.call(['make', 'version'])
    from __init__ import __version__
    version = __version__
    print('Got version string:', version)
    # v = subprocess.check_output(['git', 'describe'], text=True)
    # v = v.strip()
    # words = v.split('-')
    # if len(words) == 3:
    #     v = words[0] + '.dev' + words[1]
    # version = v
except:
    import traceback
    traceback.print_exc()
    pass

env = os.environ.copy()
#env.update(AN_GIT_REVISION=version)

class MyInstall(install):
    def run(self):
        print('MyInstall.run: calling "make -k"')
        subprocess.call(['make', '-k'], env=env)
        print('MyInstall.run: calling "make -k py"')
        subprocess.call(['make', '-k', 'py'], env=env)

        for cmd in ['make -k pyinstall',
                    'make -k install']:
            myenv = env.copy()
            dirnm = self.install_base
            if dirnm is not None:
                myenv.update(INSTALL_DIR=dirnm)
            pybase = self.install_platlib
            if pybase is not None:
                pybase = os.path.join(pybase, 'astrometry')
                myenv.update(PY_BASE_INSTALL_DIR=pybase)
            py = sys.executable
            if py is not None:
                myenv.update(PYTHON=py)
            print('Running:', cmd)
            subprocess.call(cmd, shell=True, env=myenv)
            install.run(self)

class MyBuildExt(install):
    def run(self):
        print('MyBuildExt.run: calling "make -k"')
        subprocess.call(['make', '-k'], env=env)
        print('MyBuildExt.run: calling "make -k py"')
        subprocess.call(['make', '-k', 'py'], env=env)

setup(name='astrometry',
      version=version,
      author='Astrometry.net team',
      author_email='dstndstn@gmail.com',
      url='http://astrometry.net',
      cmdclass={'install': MyInstall, 'build_ext': MyBuildExt},
      packages=['astrometry'],
      package_dir={'astrometry':''},
      )



import numpy as np

class Twrapper(object):
    def __init__(self, table, lower=True):
        '''
        *table*: astropy.table.Table object.
        '''
        self._table = table
        self._lower = lower
        self._colmap = dict()
        self._revmap = dict()

        if lower:
            for c in table.colnames:
                c2 = c.lower()
                self._colmap[c2] = c
                self._revmap[c] = c2

    def get_column_name(self, c):
        return self._colmap.get(c, c)

    def get_columns(self):
        return [self._revmap.get(c, c) for c in self._table.colnames]

    def cut(self, I):
        keep = np.zeros(len(self), bool)
        keep[I] = True
        I = np.flatnonzero(keep == False)
        self._table.remove_rows(I)

    def rename(self, k, knew):
        k2 = self.get_column_name(k)
        del self._colmap[k]
        if self._lower and k2 in self._revmap:
            del self._revmap[k2]
        self._table.rename_column(k2, knew)

    def __len__(self):
        return len(self._table)

    def __getattr__(self, k):
        k = self.get_column_name(k)
        return self._table.field(k).data

    def __setattr__(self, k, val):
        if k[0] == '_':
            return super().__setattr__(k, val)
        k2 = self.get_column_name(k)
        if k2 in self._table.colnames:
            self._table[k2] = val
        else:
            self._table.add_column(val, name=k)

    def __getitem__(self, I):
        return Twrapper(self._table[I], lower=self._lower)



from __future__ import print_function
# This file is part of libkd.
# Licensed under a 3-clause BSD style license - see LICENSE
from setuptools import setup, Extension
import os.path

from numpy import get_include
numpy_inc = get_include()

def strlist(s, split=' '):
    lst = s.split(split)
    lst = [i.strip() for i in lst]
    lst = [i for i in lst if len(i)]
    return lst

link = ' '.join([os.environ.get('LDFLAGS', ''),
                 os.environ.get('LDLIBS', ''),])
link = strlist(link)
objs = strlist(os.environ.get('SLIB', ''))
inc = strlist(os.environ.get('INC', ''), split='-I')
cflags = strlist(os.environ.get('CFLAGS', ''))

print(('link:', link))
print(('objs:', objs))
print(('inc:', inc))
print(('cflags:', cflags))

c_module = Extension('spherematch_c',
                     sources = ['pyspherematch.c'],
                     include_dirs = [numpy_inc] + inc,
                     extra_objects = objs,
                     extra_compile_args = cflags,
                     extra_link_args=link,
    )

setup(name = 'Kdtree matching in Python',
      version = '1.0',
      description = 'This package finds near neighbours in two sets of points',
      author = 'Astrometry.net (Dustin Lang)',
      author_email = 'dstndstn@gmail.com',
      url = 'http://astrometry.net',
      py_modules = [ 'spherematch' ],
      ext_modules = [c_module])




# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
from __future__ import absolute_import
import numpy as np

class ResampleError(Exception):
    pass
class OverlapError(ResampleError):
    pass
class NoOverlapError(OverlapError):
    pass
class SmallOverlapError(OverlapError):
    pass

def resample_with_wcs(targetwcs, wcs, Limages=[], L=3, spline=True,
                      splineFallback=True,
                      splineStep=25,
                      splineMargin=12,
                      table=True,
                      cinterp=True,
                      carefulWrap=False,
                      intType=np.int32):
    '''
    Returns (Yo,Xo, Yi,Xi, ims)

    Use the results like:

    target[Yo,Xo] = nearest_neighbour[Yi,Xi]
    # or
    target[Yo,Xo] = ims[i]


    raises NoOverlapError if the target and input WCSes do not
    overlap.  Raises SmallOverlapError if they do not overlap "enough"
    (as described below).

    targetwcs, wcs: duck-typed WCS objects that must have:
       - properties "imagew", "imageh"
       - methods  "r,d = pixelxy2radec(x, y)"
       -          "ok,x,y = radec2pixelxy(ra, dec)"

    The WCS functions are expected to operate in FITS pixel-indexing.

    The WCS function must support 1-d, broadcasting, vectorized
    pixel<->radec calls.

    Limages: list of images to Lanczos-interpolate at the given Lanczos order.
    If empty, just returns nearest-neighbour indices.

    L: int, lanczos order

    spline: bool: use a spline interpolator to reduce the number of
    WCS calls.

    splineFallback: bool: the spline requires a certain amount of
    spatial overlap.  With splineFallback = True, fall back to
    non-spline version.  With splineFallback = False, just raises
    SmallOverlapError.

    splineStep: approximate grid size

    table: use Lanczos3 look-up table?

    intType: type to return for integer pixel coordinates.
    (however, Yi,Xi may still be returned as int32)
    '''
    ### DEBUG
    #ps = PlotSequence('resample')
    ps = None

    H,W = int(targetwcs.imageh), int(targetwcs.imagew)
    h,w = int(      wcs.imageh), int(      wcs.imagew)

    for im in Limages:
        assert(im.shape == (h,w))

    if carefulWrap:
        # Be a bit more careful about wrap-around cases.
        # (eg, where "targewcs" wraps around, and an input image spans that
        # discontinuity).  There are still unhandled issues.  You probably also
        # want to set "spline=False" for tricky cases.
        # Start with the center pixel
        ok,x0,y0 = targetwcs.radec2pixelxy(
                *(wcs.pixelxy2radec(w/2+0.5, h/2+0.5)[-2:]))
        x1,y1 = x0,y0
        # Walk the boundary of the input image
        # and project into output image space
        # Also do a coarse grid.
        xm,ym = np.meshgrid(np.linspace(0, w-1, 20), np.linspace(0, h-1, 20))
        for x,y in [
                (np.linspace(0.,   w-1., w), np.linspace(0.,   0.,   w)),
                (np.linspace(w-1., w-1., h), np.linspace(0.,   h-1., h)),
                (np.linspace(w-1.,   0., w), np.linspace(h-1., h-1., w)),
                (np.linspace(0.,     0., h), np.linspace(h-1., 0.,   h)),
                (xm.ravel(), ym.ravel())]:
            # [-2:]: handle ok,ra,dec or ra,dec
            ok,xw,yw = targetwcs.radec2pixelxy(
                *(wcs.pixelxy2radec(x + 1., y + 1.)[-2:]))
            x0 = min(x0, min(xw))
            x1 = max(x1, max(xw))
            y0 = min(y0, min(yw))
            y1 = max(y1, max(yw))
        del xm,ym,xw,yw,x,y
        x0,y0 = int(x0),int(y0)
        x1,y1 = int(x1),int(y1)
        # # Any large breaks (eg, wrap-arounds)?
        # dd = np.hypot(np.diff(XX), np.diff(YY))
        # # jump in output pixels
        # bigjump = 100
        # I = np.flatnonzero(dd > bigjump)
        # print('dd', dd)
        # print('max', dd.max())
        # import pylab as plt
        # plt.clf()
        # plt.plot(XX, YY, 'b-')
        # plt.axis([0, 2000, 0, 1000])
        # plt.savefig('/tmp/1.png')
    else:
        # First find the approximate bbox of the input image in
        # the target image so that we don't ask for way too
        # many out-of-bounds pixels...
        XY = []
        for x,y in [(0,0), (w-1,0), (w-1,h-1), (0, h-1)]:
            # [-2:]: handle ok,ra,dec or ra,dec
            ok,xw,yw = targetwcs.radec2pixelxy(
                *(wcs.pixelxy2radec(float(x + 1), float(y + 1))[-2:]))
            XY.append((xw - 1, yw - 1))
        XY = np.array(XY)

        x0,y0 = np.rint(XY.min(axis=0))
        x1,y1 = np.rint(XY.max(axis=0))

    if spline:
        # Now we build a spline that maps "target" pixels to "input" pixels
        margin = splineMargin
        step = splineStep
        xlo = max(0, x0-margin)
        xhi = min(W-1, x1+margin)
        ylo = max(0, y0-margin)
        yhi = min(H-1, y1+margin)
        if xlo > xhi or ylo > yhi:
            raise NoOverlapError()
        nx = int(np.ceil(float(xhi - xlo) / step)) + 1
        xx = np.linspace(xlo, xhi, nx)
        ny = int(np.ceil(float(yhi - ylo) / step)) + 1
        yy = np.linspace(ylo, yhi, ny)

        if ps:
            def expand_axes():
                M = 100
                ax = plt.axis()
                plt.axis([ax[0]-M, ax[1]+M, ax[2]-M, ax[3]+M])
                plt.axis('scaled')
            plt.clf()
            plt.plot(XY[:,0], XY[:,1], 'ro')
            plt.plot(xx, np.zeros_like(xx), 'b.')
            plt.plot(np.zeros_like(yy), yy, 'c.')
            plt.plot(xx, np.zeros_like(xx)+max(yy), 'b.')
            plt.plot(max(xx) + np.zeros_like(yy), yy, 'c.')
            plt.plot([0,W,W,0,0], [0,0,H,H,0], 'k-')
            plt.title('A: Target image: bbox')
            expand_axes()
            ps.savefig()

        if (len(xx) == 0) or (len(yy) == 0):
            raise NoOverlapError()

        if (len(xx) <= 3) or (len(yy) <= 3):
            #print 'Not enough overlap between input and target WCSes'
            if splineFallback:
                spline = False
            else:
                raise SmallOverlapError()

    if spline:
        # spline inputs  -- pixel coords in the 'target' image
        #    (xx, yy)
        # spline outputs -- pixel coords in the 'input' image
        #    (XX, YY)
        # We use vectorized radec <-> pixelxy functions here

        R = targetwcs.pixelxy2radec(xx[np.newaxis,:] + 1,
                                    yy[:,np.newaxis] + 1)
        if len(R) == 3:
            ok = R[0]
            assert(np.all(ok))
        ok,XX,YY = wcs.radec2pixelxy(*(R[-2:]))
        del R
        XX -= 1.
        YY -= 1.
        assert(np.all(ok))
        del ok
        
        if ps:
            plt.clf()
            plt.plot(Xo, Yo, 'b.')
            plt.plot([0,w,w,0,0], [0,0,h,h,0], 'k-')
            plt.title('B: Input image')
            expand_axes()
            ps.savefig()
    
        import scipy.interpolate as interp
        xspline = interp.RectBivariateSpline(xx, yy, XX.T)
        yspline = interp.RectBivariateSpline(xx, yy, YY.T)
        del XX
        del YY

    else:
        margin = 0

    # Now, build the full pixel grid (in the ouput image) we want to
    # interpolate...
    ixo = np.arange(max(0, x0-margin), min(W, x1+margin+1), dtype=intType)
    iyo = np.arange(max(0, y0-margin), min(H, y1+margin+1), dtype=intType)

    if len(ixo) == 0 or len(iyo) == 0:
        raise NoOverlapError()

    okpix = True
    if spline:
        # And run the interpolator.
        # [xy]spline() does a meshgrid-like broadcast, so fxi,fyi have
        # shape n(iyo),n(ixo)
        #
        # f[xy]i: floating-point pixel coords in the input image
        fxi = xspline(ixo, iyo).T.astype(np.float32)
        fyi = yspline(ixo, iyo).T.astype(np.float32)

        if ps:
            plt.clf()
            plt.plot(ixo, np.zeros_like(ixo), 'r,')
            plt.plot(np.zeros_like(iyo), iyo, 'm,')
            plt.plot(ixo, max(iyo) + np.zeros_like(ixo), 'r,')
            plt.plot(max(ixo) + np.zeros_like(iyo), iyo, 'm,')
            plt.plot([0,W,W,0,0], [0,0,H,H,0], 'k-')
            plt.title('C: Target image; i*o')
            expand_axes()
            ps.savefig()
            plt.clf()
            plt.plot(fxi, fyi, 'r,')
            plt.plot([0,w,w,0,0], [0,0,h,h,0], 'k-')
            plt.title('D: Input image, f*i')
            expand_axes()
            ps.savefig()

    else:
        # Use 2-d broadcasting pixel <-> radec functions here.
        # This can be rather expensive, with lots of WCS calls!

        R = targetwcs.pixelxy2radec(ixo[np.newaxis,:] + 1.,
                                    iyo[:,np.newaxis] + 1.)
        if len(R) == 3:
            # ok,ra,dec
            R = R[1:]
        ok,fxi,fyi = wcs.radec2pixelxy(*R)
        #assert(np.all(ok))
        #del ok
        okpix = ok
        fxi -= 1.
        fyi -= 1.

    # i[xy]i: int coords in the input image.
    itype = intType
    if len(Limages) and cinterp:
        # the lanczos3_interpolate function below requires int32!
        itype = np.int32

    # (f + 0.5).astype(int) is often faster than round().astype(int) or rint!
    ixi = (fxi + 0.5).astype(itype)
    iyi = (fyi + 0.5).astype(itype)

    # Cut to in-bounds pixels.
    I,J = np.nonzero((ixi >= 0) * (ixi < w) * (iyi >= 0) * (iyi < h) * okpix)
    ixi = ixi[I,J]
    iyi = iyi[I,J]
    fxi = fxi[I,J]
    fyi = fyi[I,J]

    # i[xy]o: int coords in the target image.
    # These were 1-d arrays that got broadcasted
    iyo = iyo[0] + I.astype(intType)
    ixo = ixo[0] + J.astype(intType)
    del I,J

    if spline and ps:
        plt.clf()
        plt.plot(ixo, iyo, 'r,')
        plt.plot([0,W,W,0,0], [0,0,H,H,0], 'k-')
        plt.title('E: Target image; i*o')
        expand_axes()
        ps.savefig()
        plt.clf()
        plt.plot(fxi, fyi, 'r,')
        plt.plot([0,w,w,0,0], [0,0,h,h,0], 'k-')
        plt.title('F: Input image, f*i')
        expand_axes()
        ps.savefig()

    assert(np.all(ixo >= 0))
    assert(np.all(iyo >= 0))
    assert(np.all(ixo < W))
    assert(np.all(iyo < H))

    assert(np.all(ixi >= 0))
    assert(np.all(iyi >= 0))
    assert(np.all(ixi < w))
    assert(np.all(iyi < h))

    if len(Limages):
        dx = (fxi - ixi).astype(np.float32)
        dy = (fyi - iyi).astype(np.float32)
        del fxi
        del fyi

        # Lanczos interpolation.
        # number of pixels
        nn = len(ixo)
        NL = 2*L+1
        # accumulators for each input image
        laccs = [np.zeros(nn, np.float32) for im in Limages]

        if cinterp:
            from astrometry.util.util import lanczos3_interpolate
            lanczos3_interpolate(ixi, iyi, dx, dy, laccs,
                                 [lim.astype(np.float32) for lim in Limages])
        else:
            _lanczos_interpolate(L, ixi, iyi, dx, dy, laccs, Limages, table=table)
        rims = laccs
    else:
        rims = []

    return (iyo,ixo, iyi,ixi, rims)


def _lanczos_interpolate(L, ixi, iyi, dx, dy, laccs, limages,
                         table=True):
    '''
    L: int, Lanczos order
    ixi: int, 1-d numpy array, len n, x coord in input images
    iyi:     ----""----        y
    dx: float, 1-d numpy array, len n, fractional x coord
    dy:      ----""----                    y
    laccs: list of [float, 1-d numpy array, len n]: outputs
    limages list of [float, 2-d numpy array, shape h,w]: inputs
    '''
    from astrometry.util.miscutils import lanczos_filter
    lfunc = lanczos_filter
    if L == 3:
        try:
            from astrometry.util import lanczos3_filter, lanczos3_filter_table
            # 0: no rangecheck
            if table:
                lfunc = lambda nil,x,y: lanczos3_filter_table(x,y, 1)
            else:
                lfunc = lambda nil,x,y: lanczos3_filter(x,y)
        except:
            pass

    h,w = limages[0].shape
    n = len(ixi)
    # sum of lanczos terms
    fsum = np.zeros(n)
    off = np.arange(-L, L+1)
    fx = np.zeros(n, np.float32)
    fy = np.zeros(n, np.float32)
    for oy in off:
        lfunc(L, -oy + dy, fy)
        for ox in off:
            lfunc(L, -ox + dx, fx)
            for lacc,im in zip(laccs, limages):
                lacc += fx * fy * im[np.clip(iyi + oy, 0, h-1),
                                     np.clip(ixi + ox, 0, w-1)]
                fsum += fx*fy
    for lacc in laccs:
        lacc /= fsum



if __name__ == '__main__':
    import fitsio
    from astrometry.util.util import Sip,Tan
    import time
    import sys
    import pylab as plt

    from astrometry.util.util import lanczos3_filter, lanczos3_filter_table
    # x = np.linspace(-4, 4, 500)
    # L = np.zeros_like(x)
    # L2 = np.zeros(len(x), np.float32)
    # lanczos3_filter(x, L)
    # lanczos3_filter_table(x.astype(np.float32), L2, 1)
    # plt.clf()
    # plt.plot(x, L, 'r-')
    # plt.plot(x, L2, 'b-')
    # plt.savefig('l1.png')

    x = np.linspace(-3.5, 4.5, 8192).astype(np.float32)
    L1 = np.zeros_like(x)
    L2 = np.zeros_like(x)
    lanczos3_filter(x, L1)
    lanczos3_filter_table(x, L2, 1)
    print('L2 - L1 RMS:', np.sqrt(np.mean((L2-L1)**2)))
    
    if True:
        ra,dec = 0.,0.,
        pixscale = 1e-3
        W,H = 10,1

        cowcs = Tan(ra, dec, (W+1)/2., (H+1)/2.,
                    -pixscale, 0., 0., pixscale, W, H)
        dx,dy = 0.25, 0.
        wcs = Tan(ra, dec, (W+1)/2. + dx, (H+1)/2. + dy,
                  -pixscale, 0., 0., pixscale, W, H)

        pix = np.zeros((H,W), np.float32)
        pix[0,W//2] = 1.
        
        Yo,Xo,Yi,Xi,(cpix,) = resample_with_wcs(cowcs, wcs, [pix], 3)
        print('C', cpix)
        Yo2,Xo2,Yi2,Xi2,(pypix,) = resample_with_wcs(cowcs, wcs, [pix], 3, cinterp=False, table=False)
        print('Py', pypix)

        print('RMS', np.sqrt(np.mean((cpix - pypix)**2)))
        
        sys.exit(0)
        
        
    if True:
        ra,dec = 219.577111, 54.52
        pixscale = 2.75 / 3600.
        W,H = 10,10
        cowcs = Tan(ra, dec, (W+1)/2., (H+1)/2.,
                    -pixscale, 0., 0., pixscale, W, H)

        for i,(dx,dy) in enumerate([(0.01, 0.02),
                                    (0.1, 0.0),
                                    (0.2, 0.0),
                                    (0.3, 0.0),
                                    (0.4, 0.0),
                                    (0.5, 0.0),
                                    (0.6, 0.0),
                                    (0.7, 0.0),
                                    (0.8, 0.0),
                                    ]):
            wcs = Tan(ra, dec, (W+1)/2. + dx, (H+1)/2. + dy,
                      -pixscale, 0., 0., pixscale, W, H)
            pix = np.zeros((H,W), np.float32)
            pix[H/2, :] = 1.
            pix[:, W/2] = 1.
    
            Yo,Xo,Yi,Xi,(cpix,) = resample_with_wcs(cowcs, wcs, [pix], 3)
            Yo2,Xo2,Yi2,Xi2,(pypix,) = resample_with_wcs(cowcs, wcs, [pix], 3, cinterp=False)
            cim = np.zeros((H,W))
            cim[Yo,Xo] = cpix
            pyim = np.zeros((H,W))
            pyim[Yo2,Xo2] = pypix

            plt.clf()
            plt.plot(cim[0,:], 'b-', alpha=0.5)
            plt.plot(cim[H/4,:], 'c-', alpha=0.5)
            plt.plot(pyim[0,:], 'r-', alpha=0.5)
            plt.plot(pyim[H/4,:], 'm-', alpha=0.5)
            plt.plot(1000. * (cim[0,:] - pyim[0,:]), 'k-', alpha=0.5)
            plt.savefig('p2-%02i.png' % i)
        sys.exit(0)
    
    ra,dec = 219.577111, 54.52
    pixscale = 2.75 / 3600.
    #W,H = 2048, 2048
    W,H = 512, 512
    #W,H = 100,100
    cowcs = Tan(ra, dec, (W+1)/2., (H+1)/2.,
                -pixscale, 0., 0., pixscale, W, H)
    cowcs.write_to('co.wcs')
    
    if True:
        #intfn = '05579a167-w1-int-1b.fits'
        intfn = 'wise-frames/9a/05579a/167/05579a167-w1-int-1b.fits'
        wcs = Sip(intfn)
        pix = fitsio.read(intfn)
        pix[np.logical_not(np.isfinite(pix))] = 0.
        print('pix', pix.shape, pix.dtype)

    
    for i in range(5):
        t0 = time.process_time()
        Yo,Xo,Yi,Xi,ims = resample_with_wcs(cowcs, wcs, [pix], 3)
        t1 = time.process_time() - t0
        print('C resampling took', t1)

    t0 = time.process_time()
    Yo2,Xo2,Yi2,Xi2,ims2 = resample_with_wcs(cowcs, wcs, [pix], 3, cinterp=False, table=False)
    t2 = time.process_time() - t0
    print('py resampling took', t2)
    
    out = np.zeros((H,W))
    out[Yo,Xo] = ims[0]
    fitsio.write('resampled-c.fits', out, clobber=True)
    cout = out
    
    out = np.zeros((H,W))
    out[Yo,Xo] = ims2[0]
    fitsio.write('resampled-py.fits', out, clobber=True)
    pyout = out

    plt.clf()
    plt.imshow(cout, interpolation='nearest', origin='lower')
    plt.colorbar()
    plt.savefig('c.png')
    plt.clf()
    plt.imshow(pyout, interpolation='nearest', origin='lower')
    plt.colorbar()
    plt.savefig('py.png')

    plt.clf()
    plt.imshow(cout - pyout, interpolation='nearest', origin='lower')
    plt.colorbar()
    plt.savefig('diff.png')

    print('Max diff:', np.abs(cout - pyout).max())




#! /usr/bin/env python3
# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
import os
import sys
import logging
from optparse import OptionParser

import numpy as np
from astrometry.util.fits import fits_table

def uniformize(infile, outfile, n, xcol='X', ycol='Y', ext=1, **kwargs):
    T = fits_table(infile, lower=False)
    if len(T) == 0:
        print('No sources')
        T.writeto(outfile, header=T.get_header())
        return
    x = T.get(xcol)
    y = T.get(ycol)
    # use IMAGEW,H, or compute bounds?
    #  #$)(*&%^ NaNs in LSST source positions.  Seriously, WTF!
    I = np.logical_and(np.isfinite(x), np.isfinite(y))
    if not all(I):
        print('%i source positions are not finite.' % np.sum(np.logical_not(I)))
        x = x[I]
        y = y[I]
        T.cut(I)
    
    W = max(x) - min(x)
    H = max(y) - min(y)
    if W == 0 or H == 0:
        print('Area of the rectangle enclosing all image sources: %i x %i' % (W,H))
        T.writeto(outfile)
        return
    NX = int(max(1, np.round(W / np.sqrt(W*H / float(n)))))
    NY = int(max(1, np.round(n / float(NX))))
    print('Uniformizing into %i x %i bins' % (NX, NY))
    print('Image bounds: x [%g,%g], y [%g,%g]' % (min(x),max(x),min(y),max(y)))

    ix = (np.clip(np.floor((x - min(x)) / float(W) * NX), 0, NX-1)).astype(int)
    iy = (np.clip(np.floor((y - min(y)) / float(H) * NY), 0, NY-1)).astype(int)
    assert(np.all(ix >= 0))
    assert(np.all(ix < NX))
    assert(np.all(iy >= 0))
    assert(np.all(iy < NY))
    I = iy * NX + ix
    assert(np.all(I >= 0))
    assert(np.all(I < NX*NY))
    bins = [[] for i in range(NX*NY)]
    for j,i in enumerate(I):
        bins[int(i)].append(j)
    maxlen = max([len(b) for b in bins])
    J = []
    for i in range(maxlen):
        thisrow = []
        for b in bins:
            if i >= len(b):
                continue
            thisrow.append(b[i])
        thisrow.sort()
        J += thisrow
    J = np.array(J)
    #header.add_history('This xylist was filtered by the "uniformize.py" program')
    T.cut(J)
    T.writeto(outfile, header=T.get_header())
    return 0

def main():
    parser = OptionParser(usage='%prog [options] <input-xylist> <output-xylist>')

    parser.add_option('-X', dest='xcol', help='Name of X column in input table',
                      default='X')
    parser.add_option('-Y', dest='ycol', help='Name of Y column in input table',
                      default='Y')
    parser.add_option('-n', dest='n', type='int', default=10,
                      help='Number of boxes, approximately')
    parser.add_option('-e', dest='ext', type=int, help='FITS extension to read',
                      default=1)
    (opt, args) = parser.parse_args()
    
    if len(args) != 2:
        parser.print_help()
        print()
        print('Got arguments:', args)
        sys.exit(-1)

    infile = args[0]
    outfile = args[1]
    return uniformize(infile, outfile, opt.n, xcol=opt.xcol, ycol=opt.ycol,
                      ext=opt.ext)

if __name__ == '__main__':
    sys.exit(main())




# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function

import multiprocessing

class FakeAsyncResult(object):
    def __init__(self, X):
        self.X = X
    def wait(self, *a):
        pass
    def get(self, *a):
        return self.X
    def ready(self):
        return True
    def successful(self):
        return True

class funcwrapper(object):
    def __init__(self, func):
        self.func = func
    def __call__(self, *X):
        #print 'Trying to call', self.func
        #print 'with args', X
        try:
            return self.func(*X)
        except:
            import traceback
            print('Exception while calling your function:')
            print('  params:', X)
            print('  exception:')
            traceback.print_exc()
            raise

class memberfuncwrapper(object):
    def __init__(self, obj, funcname):
        self.obj = obj
        self.funcname = funcname
    def __call__(self, *X):
        func = self.obj.getattr(self.funcname)
        #print 'Trying to call', self.func
        #print 'with args', X
        try:
            return func(self.obj, *X)
        except:
            import traceback
            print('Exception while calling your function:')
            print('  object:', self.obj)
            print('  member function:', self.funcname)
            print('  ', func)
            print('  params:', X)
            print('  exception:')
            traceback.print_exc()
            raise



class multiproc(object):
    def __init__(self, nthreads=1, init=None, initargs=[],
                 map_chunksize=1, pool=None, wrap_all=False):
        self.wrap_all = wrap_all
        if pool is not None:
            self.pool = pool
            self.applyfunc = self.pool.apply_async
        else:
            if nthreads == 1:
                self.pool = None
                # self.map = map
                self.applyfunc = lambda f,a,k: f(*a, **k)
                if init is not None:
                    init(*initargs)
            else:
                self.pool = multiprocessing.Pool(nthreads, init, initargs)
                # self.map = self.pool.map
                self.applyfunc = self.pool.apply_async
        self.async_results = []
        self.map_chunksize = map_chunksize

    def map(self, f, args, chunksize=None, wrap=False):
        cs = chunksize
        if cs is None:
            cs = self.map_chunksize
        if self.pool:
            if wrap or self.wrap_all:
                f = funcwrapper(f)
            #print 'pool.map: f', f
            #print 'args', args
            #print 'cs', cs
            return self.pool.map(f, args, cs)
        return list(map(f, args))

    def map_async(self, func, iterable, wrap=False):
        if self.pool is None:
            return FakeAsyncResult(map(func, iterable))
        if wrap or self.wrap_all:
            return self.pool.map_async(funcwrapper(func), iterable)
        return self.pool.map_async(func, iterable)

    def imap(self, func, iterable, chunksize=None, wrap=False):
        cs = chunksize
        if cs is None:
            cs = self.map_chunksize
        if self.pool is None:
            import itertools
            if 'imap' in dir(itertools):
                # py2
                return itertools.imap(func, iterable)
            else:
                # py3
                return map(func, iterable)
        if wrap or self.wrap_all:
            func = funcwrapper(func)
        return self.pool.imap(func, iterable, chunksize=cs)

    def imap_unordered(self, func, iterable, chunksize=None, wrap=False):
        cs = chunksize
        if cs is None:
            cs = self.map_chunksize
        if self.pool is None:
            import itertools
            if 'imap' in dir(itertools):
                # py2
                return itertools.imap(func, iterable)
            else:
                # py3
                return map(func, iterable)
        if wrap or self.wrap_all:
            func = funcwrapper(func)
        return self.pool.imap_unordered(func, iterable, chunksize=cs)
    
    def apply(self, f, args, wrap=False, kwargs={}):
        if self.pool is None:
            return FakeAsyncResult(f(*args, **kwargs))
        if wrap:
            f = funcwrapper(f)
        res = self.applyfunc(f, args, kwargs)
        self.async_results.append(res)
        return res

    def waitforall(self):
        print('Waiting for async results to finish...')
        for r in self.async_results:
            print('  waiting for', r)
            r.wait()
        print('all done')
        self.async_results = []

    def close(self):
        if self.pool is not None:
            self.pool.close()
            self.pool = None

